import io
from typing import (
    Any,
    List,
    Optional,
    Tuple,
    TYPE_CHECKING,
    Union,
)

from ._types import (
    Annotated,
    AssertionParameter,
    Negate,
    NEGATE_DEFAULT,
    OptionalXmlFloat,
    OptionalXmlInt,
    OutputBytes,
    XmlFloat,
    XmlInt,
)
from ._util import _assert_number

try:
    import numpy
except ImportError:
    pass
try:
    from PIL import Image
except ImportError:
    pass
try:
    import tifffile
except ImportError:
    pass

if TYPE_CHECKING:
    import numpy.typing


JSON_STRICT_NUMBER = "typing.Union[StrictInt, StrictFloat]"
JSON_OPTIONAL_STRICT_NUMBER = f"typing.Optional[{JSON_STRICT_NUMBER}]"


Width = Annotated[
    OptionalXmlInt,
    AssertionParameter(
        "Expected width of the image (in pixels).",
        json_type="typing.Optional[StrictInt]",
        xml_type="xs:nonNegativeInteger",
        validators=["check_non_negative_if_set"],
    ),
]
Height = Annotated[
    OptionalXmlInt,
    AssertionParameter(
        "Expected height of the image (in pixels).",
        json_type="typing.Optional[StrictInt]",
        xml_type="xs:nonNegativeInteger",
        validators=["check_non_negative_if_set"],
    ),
]
Channels = Annotated[
    OptionalXmlInt,
    AssertionParameter(
        "Expected number of channels of the image.",
        json_type="typing.Optional[StrictInt]",
        xml_type="xs:nonNegativeInteger",
        validators=["check_non_negative_if_set"],
    ),
]
Depth = Annotated[
    OptionalXmlInt,
    AssertionParameter(
        "Expected depth of the image (number of slices).",
        json_type="typing.Optional[StrictInt]",
        xml_type="xs:nonNegativeInteger",
        validators=["check_non_negative_if_set"],
    ),
]
Frames = Annotated[
    OptionalXmlInt,
    AssertionParameter(
        "Expected number of frames in the image sequence (number of time steps).",
        json_type="typing.Optional[StrictInt]",
        xml_type="xs:nonNegativeInteger",
        validators=["check_non_negative_if_set"],
    ),
]
WidthDelta = Annotated[
    XmlInt,
    AssertionParameter(
        "Maximum allowed difference of the image width (in pixels, default is 0). The observed width has to be in the range ``value +- delta``.",
        json_type="StrictInt",
        xml_type="xs:nonNegativeInteger",
        validators=["check_non_negative_if_set"],
    ),
]
WidthMin = Annotated[
    OptionalXmlInt,
    AssertionParameter(
        "Minimum allowed width of the image (in pixels).",
        json_type="typing.Optional[StrictInt]",
        xml_type="xs:nonNegativeInteger",
        validators=["check_non_negative_if_set"],
    ),
]
WidthMax = Annotated[
    OptionalXmlInt,
    AssertionParameter(
        "Maximum allowed width of the image (in pixels).",
        json_type="typing.Optional[StrictInt]",
        xml_type="xs:nonNegativeInteger",
        validators=["check_non_negative_if_set"],
    ),
]
HeightDelta = Annotated[
    XmlInt,
    AssertionParameter(
        "Maximum allowed difference of the image height (in pixels, default is 0). The observed height has to be in the range ``value +- delta``.",
        json_type="StrictInt",
        xml_type="xs:nonNegativeInteger",
        validators=["check_non_negative_if_set"],
    ),
]
HeightMin = Annotated[
    OptionalXmlInt,
    AssertionParameter(
        "Minimum allowed height of the image (in pixels).",
        json_type="typing.Optional[StrictInt]",
        xml_type="xs:nonNegativeInteger",
        validators=["check_non_negative_if_set"],
    ),
]
HeightMax = Annotated[
    OptionalXmlInt,
    AssertionParameter(
        "Maximum allowed height of the image (in pixels).",
        json_type="typing.Optional[StrictInt]",
        xml_type="xs:nonNegativeInteger",
        validators=["check_non_negative_if_set"],
    ),
]
ChannelsDelta = Annotated[
    XmlInt,
    AssertionParameter(
        "Maximum allowed difference of the number of channels (default is 0). The observed number of channels has to be in the range ``value +- delta``.",
        json_type="StrictInt",
        xml_type="xs:nonNegativeInteger",
        validators=["check_non_negative_if_set"],
    ),
]
ChannelsMin = Annotated[
    OptionalXmlInt,
    AssertionParameter(
        "Minimum allowed number of channels.",
        json_type="typing.Optional[StrictInt]",
        xml_type="xs:nonNegativeInteger",
        validators=["check_non_negative_if_set"],
    ),
]
ChannelsMax = Annotated[
    OptionalXmlInt,
    AssertionParameter(
        "Maximum allowed number of channels.",
        json_type="typing.Optional[StrictInt]",
        xml_type="xs:nonNegativeInteger",
        validators=["check_non_negative_if_set"],
    ),
]
DepthDelta = Annotated[
    XmlInt,
    AssertionParameter(
        "Maximum allowed difference of the image depth (number of slices, default is 0). The observed depth has to be in the range ``value +- delta``.",
        json_type="StrictInt",
        xml_type="xs:nonNegativeInteger",
        validators=["check_non_negative_if_set"],
    ),
]
DepthMin = Annotated[
    OptionalXmlInt,
    AssertionParameter(
        "Minimum allowed depth of the image (number of slices).",
        json_type="typing.Optional[StrictInt]",
        xml_type="xs:nonNegativeInteger",
        validators=["check_non_negative_if_set"],
    ),
]
DepthMax = Annotated[
    OptionalXmlInt,
    AssertionParameter(
        "Maximum allowed depth of the image (number of slices).",
        json_type="typing.Optional[StrictInt]",
        xml_type="xs:nonNegativeInteger",
        validators=["check_non_negative_if_set"],
    ),
]
FramesDelta = Annotated[
    XmlInt,
    AssertionParameter(
        "Maximum allowed difference of the number of frames in the image sequence (number of time steps, default is 0). The observed number of frames has to be in the range ``value +- delta``.",
        json_type="StrictInt",
        xml_type="xs:nonNegativeInteger",
        validators=["check_non_negative_if_set"],
    ),
]
FramesMin = Annotated[
    OptionalXmlInt,
    AssertionParameter(
        "Minimum allowed number of frames in the image sequence (number of time steps).",
        json_type="typing.Optional[StrictInt]",
        xml_type="xs:nonNegativeInteger",
        validators=["check_non_negative_if_set"],
    ),
]
FramesMax = Annotated[
    OptionalXmlInt,
    AssertionParameter(
        "Maximum allowed number of frames in the image sequence (number of time steps).",
        json_type="typing.Optional[StrictInt]",
        xml_type="xs:nonNegativeInteger",
        validators=["check_non_negative_if_set"],
    ),
]
MeanIntensity = Annotated[
    OptionalXmlFloat,
    AssertionParameter("The required mean value of the image intensities.", json_type=JSON_OPTIONAL_STRICT_NUMBER),
]
MeanIntensityEps = Annotated[
    XmlFloat,
    AssertionParameter(
        "The absolute tolerance to be used for ``value`` (defaults to ``0.01``). The observed mean value of the image intensities has to be in the range ``value +- eps``.",
        json_type=JSON_STRICT_NUMBER,
        validators=["check_non_negative_if_set"],
    ),
]
MeanIntensityMin = Annotated[
    OptionalXmlFloat,
    AssertionParameter(
        "A lower bound of the required mean value of the image intensities.", json_type=JSON_OPTIONAL_STRICT_NUMBER
    ),
]
MeanIntensityMax = Annotated[
    OptionalXmlFloat,
    AssertionParameter(
        "An upper bound of the required mean value of the image intensities.", json_type=JSON_OPTIONAL_STRICT_NUMBER
    ),
]
NumLabels = Annotated[
    OptionalXmlInt,
    AssertionParameter(
        "Expected number of labels.",
        json_type="typing.Optional[StrictInt]",
        xml_type="xs:nonNegativeInteger",
        validators=["check_non_negative_if_set"],
    ),
]
NumLabelsDelta = Annotated[
    XmlInt,
    AssertionParameter(
        "Maximum allowed difference of the number of labels (default is 0). The observed number of labels has to be in the range ``value +- delta``.",
        json_type="StrictInt",
        xml_type="xs:nonNegativeInteger",
        validators=["check_non_negative_if_set"],
    ),
]
NumLabelsMin = Annotated[
    OptionalXmlInt,
    AssertionParameter(
        "Minimum allowed number of labels.",
        json_type="typing.Optional[StrictInt]",
        xml_type="xs:nonNegativeInteger",
        validators=["check_non_negative_if_set"],
    ),
]
NumLabelsMax = Annotated[
    OptionalXmlInt,
    AssertionParameter(
        "Maximum allowed number of labels.",
        json_type="typing.Optional[StrictInt]",
        xml_type="xs:nonNegativeInteger",
        validators=["check_non_negative_if_set"],
    ),
]

Channel = Annotated[
    OptionalXmlInt,
    AssertionParameter(
        "Restricts the assertion to a specific channel of the image (where ``0`` corresponds to the first image channel).",
        json_type="typing.Optional[StrictInt]",
    ),
]
Slice = Annotated[
    OptionalXmlInt,
    AssertionParameter(
        "Restricts the assertion to a specific slice of the image (where ``0`` corresponds to the first image slice).",
        json_type="typing.Optional[StrictInt]",
    ),
]
Frame = Annotated[
    OptionalXmlInt,
    AssertionParameter(
        "Restricts the assertion to a specific frame of the image sequence (where ``0`` corresponds to the first image frame).",
        json_type="typing.Optional[StrictInt]",
    ),
]
CenterOfMass = Annotated[
    str,
    AssertionParameter(
        "The required center of mass of the image intensities (horizontal and vertical coordinate, separated by a comma).",
        validators=["check_center_of_mass"],
    ),
]
CenterOfMassEps = Annotated[
    XmlFloat,
    AssertionParameter(
        "The maximum allowed Euclidean distance to the required center of mass (defaults to ``0.01``).",
        json_type=JSON_STRICT_NUMBER,
        validators=["check_non_negative_if_set"],
    ),
]
Labels = Annotated[
    Optional[Union[str, List[Union[float, int]]]],
    AssertionParameter(
        "List of labels, separated by a comma. Labels *not* on this list will be excluded from consideration. Cannot be used in combination with ``exclude_labels``.",
        xml_type="xs:string",
        json_type=f"typing.Optional[typing.List[{JSON_STRICT_NUMBER}]]",
    ),
]
ExcludeLabels = Annotated[
    Optional[Union[str, List[Union[float, int]]]],
    AssertionParameter(
        "List of labels to be excluded from consideration, separated by a comma. The primary usage of this attribute is to exclude the background of a label image. Cannot be used in combination with ``labels``.",
        xml_type="xs:string",
        json_type=f"typing.Optional[typing.List[{JSON_STRICT_NUMBER}]]",
    ),
]
MeanObjectSize = Annotated[
    OptionalXmlFloat,
    AssertionParameter(
        "The required mean size of the uniquely labeled objects.",
        xml_type="xs:float",
        json_type=JSON_OPTIONAL_STRICT_NUMBER,
        validators=["check_non_negative_if_set"],
    ),
]
MeanObjectSizeEps = Annotated[
    XmlFloat,
    AssertionParameter(
        "The absolute tolerance to be used for ``value`` (defaults to ``0.01``). The observed mean size of the uniquely labeled objects has to be in the range ``value +- eps``.",
        xml_type="xs:float",
        json_type=JSON_STRICT_NUMBER,
        validators=["check_non_negative_if_set"],
    ),
]
MeanObjectSizeMin = Annotated[
    OptionalXmlFloat,
    AssertionParameter(
        "A lower bound of the required mean size of the uniquely labeled objects.",
        xml_type="xs:float",
        json_type=JSON_OPTIONAL_STRICT_NUMBER,
        validators=["check_non_negative_if_set"],
    ),
]
MeanObjectSizeMax = Annotated[
    OptionalXmlFloat,
    AssertionParameter(
        "An upper bound of the required mean size of the uniquely labeled objects.",
        xml_type="xs:float",
        json_type=JSON_OPTIONAL_STRICT_NUMBER,
        validators=["check_non_negative_if_set"],
    ),
]


def _assert_float(
    actual: float,
    label: str,
    tolerance: Union[float, str],
    expected: Optional[Union[float, str]] = None,
    range_min: Optional[Union[float, str]] = None,
    range_max: Optional[Union[float, str]] = None,
) -> None:

    # Perform `tolerance` based check.
    if expected is not None:
        assert abs(actual - float(expected)) <= float(
            tolerance
        ), f"Wrong {label}: {actual} (expected {expected} ±{tolerance})"

    # Perform `range_min` based check.
    if range_min is not None:
        assert actual >= float(range_min), f"Wrong {label}: {actual} (must be {range_min} or larger)"

    # Perform `range_max` based check.
    if range_max is not None:
        assert actual <= float(range_max), f"Wrong {label}: {actual} (must be {range_max} or smaller)"


def assert_has_image_width(
    output_bytes: OutputBytes,
    width: Width = None,
    delta: WidthDelta = 0,
    min: WidthMin = None,
    max: WidthMax = None,
    negate: Negate = NEGATE_DEFAULT,
) -> None:
    """
    Asserts the output is an image and has a specific width (in pixels).

    The width is plus/minus ``delta`` (e.g., ``<has_image_width width="512" delta="2" />``).
    Alternatively the range of the expected width can be specified by ``min`` and/or ``max``.
    """
    im_arr = _get_image(output_bytes)
    _assert_number(
        im_arr.shape[3],  # Image axes are normalized like "TZYXC"
        width,
        delta,
        min,
        max,
        negate,
        "{expected} width {n}+-{delta}",
        "{expected} width to be in [{min}:{max}]",
    )


def assert_has_image_height(
    output_bytes: bytes,
    height: Height = None,
    delta: HeightDelta = 0,
    min: HeightMin = None,
    max: HeightMax = None,
    negate: Negate = NEGATE_DEFAULT,
) -> None:
    """
    Asserts the output is an image and has a specific height (in pixels).

    The height is plus/minus ``delta`` (e.g., ``<has_image_height height="512" delta="2" />``).
    Alternatively the range of the expected height can be specified by ``min`` and/or ``max``.
    """
    im_arr = _get_image(output_bytes)
    _assert_number(
        im_arr.shape[2],  # Image axes are normalized like "TZYXC"
        height,
        delta,
        min,
        max,
        negate,
        "{expected} height {n}+-{delta}",
        "{expected} height to be in [{min}:{max}]",
    )


def assert_has_image_channels(
    output_bytes: bytes,
    channels: Channels = None,
    delta: ChannelsDelta = 0,
    min: ChannelsMin = None,
    max: ChannelsMax = None,
    negate: Negate = NEGATE_DEFAULT,
) -> None:
    """
    Asserts the output is an image and has a specific number of channels.

    The number of channels is plus/minus ``delta`` (e.g., ``<has_image_channels channels="3" />``).

    Alternatively the range of the expected number of channels can be specified by ``min`` and/or ``max``.
    """
    im_arr = _get_image(output_bytes)
    _assert_number(
        im_arr.shape[-1],  # Image axes are normalized like "TZYXC"
        channels,
        delta,
        min,
        max,
        negate,
        "{expected} image channels {n}+-{delta}",
        "{expected} image channels to be in [{min}:{max}]",
    )


def assert_has_image_depth(
    output_bytes: OutputBytes,
    depth: Depth = None,
    delta: DepthDelta = 0,
    min: DepthMin = None,
    max: DepthMax = None,
    negate: Negate = NEGATE_DEFAULT,
) -> None:
    """
    Asserts the output is an image and has a specific depth (number of slices).

    The depth is plus/minus ``delta`` (e.g., ``<has_image_depth depth="512" delta="2" />``).
    Alternatively the range of the expected depth can be specified by ``min`` and/or ``max``.
    """
    im_arr = _get_image(output_bytes)
    _assert_number(
        im_arr.shape[1],  # Image axes are normalized like "TZYXC"
        depth,
        delta,
        min,
        max,
        negate,
        "{expected} depth {n}+-{delta}",
        "{expected} depth to be in [{min}:{max}]",
    )


def assert_has_image_frames(
    output_bytes: OutputBytes,
    frames: Frames = None,
    delta: FramesDelta = 0,
    min: FramesMin = None,
    max: FramesMax = None,
    negate: Negate = NEGATE_DEFAULT,
) -> None:
    """
    Asserts the output is an image and has a specific number of frames (number of time steps).

    The number of frames is plus/minus ``delta`` (e.g., ``<has_image_frames depth="512" delta="2" />``).
    Alternatively the range of the expected number of frames can be specified by ``min`` and/or ``max``.
    """
    im_arr = _get_image(output_bytes)
    _assert_number(
        im_arr.shape[0],  # Image axes are normalized like "TZYXC"
        frames,
        delta,
        min,
        max,
        negate,
        "{expected} frames {n}+-{delta}",
        "{expected} frames to be in [{min}:{max}]",
    )


def _compute_center_of_mass(im_arr: "numpy.typing.NDArray") -> Tuple[float, float]:
    im_arr_yx = im_arr.sum(axis=(0, 1, 4))  # Image axes are normalized like "TZYXC"
    im_arr_yx = numpy.abs(im_arr_yx)
    if im_arr_yx.sum() == 0:
        return (numpy.nan, numpy.nan)
    im_arr_yx = im_arr_yx / im_arr_yx.sum()
    yy, xx = numpy.indices(im_arr_yx.shape)
    return (im_arr_yx * xx).sum(), (im_arr_yx * yy).sum()


def _move_char(s: str, pos_src: int, pos_dst: int) -> str:
    s_list = list(s)
    c = s_list.pop(pos_src)
    if pos_dst < 0:
        pos_dst = len(s_list) + pos_dst + 1
    s_list.insert(pos_dst, c)
    return "".join(s_list)


def _swap_char(s: str, pos1: int, pos2: int) -> str:
    s_list = list(s)
    s_list[pos1], s_list[pos2] = s_list[pos2], s_list[pos1]
    return "".join(s_list)


def _get_image(
    output_bytes: bytes,
    channel: Optional[Union[int, str]] = None,
    slice: Optional[Union[int, str]] = None,
    frame: Optional[Union[int, str]] = None,
) -> "numpy.typing.NDArray":
    """
    Returns the output image with the axes ``TZYXC``, optionally restricted to a specific `channel`, `slice`,
    `frame`, or a combination thereof.

    The function tries to read the image using tifffile and Pillow. The image axes are normalized like ``TZYXC``,
    treating sample axis ``S`` as an alias for the channel axis ``C``. For images which cannot be read by tifffile,
    two- and three-dimensional data is supported. Two-dimensional images are assumed to be in ``YX`` axes order,
    and three-dimensional images are assumed to be in ``YXC`` axes order.
    """
    buf = io.BytesIO(output_bytes)

    # Try reading with tifffile first. It fails if the file is not a TIFF.
    try:
        with tifffile.TiffFile(buf) as im_file:
            assert len(im_file.series) == 1, f"Image has unsupported number of series: {len(im_file.series)}"
            im_axes = im_file.series[0].axes

            # Verify that the image format is supported
            assert (
                frozenset("YX") <= frozenset(im_axes) <= frozenset("TZYXCS")
            ), f"Image has unsupported axes: {im_axes}"

            # Treat sample axis "S" as channel axis "C" and fail if both are present
            assert (
                "C" not in im_axes or "S" not in im_axes
            ), f"Image has sample and channel axes which is not supported: {im_axes}"
            im_axes = im_axes.replace("S", "C")

            # Read the image data
            im_arr = im_file.asarray()

            # Step 1. In the three steps below, the optional axes are added, of if they arent't there yet:

            # (1.1) Append "C" axis if not present yet
            if im_axes.find("C") == -1:
                im_arr = im_arr[..., None]
                im_axes += "C"

            # (1.2) Append "Z" axis if not present yet
            if im_axes.find("Z") == -1:
                im_arr = im_arr[..., None]
                im_axes += "Z"

            # (1.3) Append "T" axis if not present yet
            if im_axes.find("T") == -1:
                im_arr = im_arr[..., None]
                im_axes += "T"

            # Step 2. All supported axes are there now. Normalize the order of the axes:

            # (2.1) Normalize order of axes "Y" and "X"
            ypos = im_axes.find("Y")
            xpos = im_axes.find("X")
            if ypos > xpos:
                im_arr = im_arr.swapaxes(ypos, xpos)
                im_axes = _swap_char(im_axes, xpos, ypos)

            # (2.2) Normalize the position of the "C" axis (should be last)
            cpos = im_axes.find("C")
            if cpos < len(im_axes) - 1:
                im_arr = numpy.moveaxis(im_arr, cpos, -1)
                im_axes = _move_char(im_axes, cpos, -1)

            # (2.3) Normalize the position of the "T" axis (should be first)
            tpos = im_axes.find("T")
            if tpos != 0:
                im_arr = numpy.moveaxis(im_arr, tpos, 0)
                im_axes = _move_char(im_axes, tpos, 0)

            # (2.4) Normalize the position of the "Z" axis (should be second)
            zpos = im_axes.find("Z")
            if zpos != 1:
                im_arr = numpy.moveaxis(im_arr, zpos, 1)
                im_axes = _move_char(im_axes, zpos, 1)

            # Verify that the normalizations were successful
            assert im_axes == "TZYXC", f"Image axis normalization failed: {im_axes}"

    # If tifffile failed, then the file is not a tifffile. In that case, try with Pillow.
    except tifffile.TiffFileError:
        buf.seek(0)
        with Image.open(buf) as im:
            im_arr = numpy.array(im)

            # Verify that the image format is supported
            assert im_arr.ndim in (2, 3), f"Image has unsupported dimension: {im_arr.ndim}"

            # Normalize the axes
            if im_arr.ndim == 2:  # Append "C" axis if not present yet
                im_arr = im_arr[..., None]
            im_arr = im_arr[None, None, ...]  # Prepend "T" and "Z" axes

            # Verify that the normalizations were successful
            assert im_arr.ndim == 5, "Image axis normalization failed"

    # Select the specified channel (if any).
    if channel is not None:
        im_arr = im_arr[..., [int(channel)]]

    # Select the specified slice (if any).
    if slice is not None:
        im_arr = im_arr[:, [int(slice)], ...]

    # Select the specified frame (if any).
    if frame is not None:
        im_arr = im_arr[[int(frame)], ...]

    # Return the image
    return im_arr


def assert_has_image_mean_intensity(
    output_bytes: OutputBytes,
    channel: Channel = None,
    slice: Slice = None,
    frame: Frame = None,
    mean_intensity: MeanIntensity = None,
    eps: MeanIntensityEps = 0.01,
    min: MeanIntensityMin = None,
    max: MeanIntensityMax = None,
) -> None:
    """
    Asserts the output is an image and has a specific mean intensity value.

    The mean intensity value is plus/minus ``eps`` (e.g., ``<has_image_mean_intensity mean_intensity="0.83" />``).
    Alternatively the range of the expected mean intensity value can be specified by ``min`` and/or ``max``.
    """
    im_arr = _get_image(output_bytes, channel, slice, frame)
    _assert_float(
        actual=im_arr.mean(),
        label="mean intensity",
        tolerance=eps,
        expected=mean_intensity,
        range_min=min,
        range_max=max,
    )


def assert_has_image_center_of_mass(
    output_bytes: OutputBytes,
    center_of_mass: CenterOfMass,
    channel: Channel = None,
    slice: Slice = None,
    frame: Frame = None,
    eps: CenterOfMassEps = 0.01,
) -> None:
    """
    Asserts the specified output is an image and has the specified center of mass.

    Asserts the output is an image and has a specific center of mass,
    or has an Euclidean distance of ``eps`` or less to that point (e.g.,
    ``<has_image_center_of_mass center_of_mass="511.07, 223.34" />``).
    """
    im_arr = _get_image(output_bytes, channel, slice, frame)
    center_of_mass_parts = [c.strip() for c in center_of_mass.split(",")]
    assert len(center_of_mass_parts) == 2
    center_of_mass_tuple = (float(center_of_mass_parts[0]), float(center_of_mass_parts[1]))
    assert len(center_of_mass_tuple) == 2, "center_of_mass must have two components"
    actual_center_of_mass = _compute_center_of_mass(im_arr)
    distance = numpy.linalg.norm(numpy.subtract(center_of_mass_tuple, actual_center_of_mass))
    assert distance <= float(
        eps
    ), f"Wrong center of mass: {actual_center_of_mass} (expected {center_of_mass}, distance: {distance}, eps: {eps})"


def _get_image_labels(
    output_bytes: bytes,
    channel: Optional[Union[int, str]] = None,
    slice: Optional[Union[int, str]] = None,
    frame: Optional[Union[int, str]] = None,
    labels: Labels = None,
    exclude_labels: ExcludeLabels = None,
) -> Tuple["numpy.typing.NDArray", List[Any]]:
    """
    Determines the unique labels in the output image or a specific channel.
    """
    assert labels is None or exclude_labels is None
    im_arr = _get_image(output_bytes, channel, slice, frame)

    def cast_label(label: str) -> Union[float, int]:
        label = label.strip()
        if numpy.issubdtype(im_arr.dtype, numpy.integer):
            return int(label)
        if numpy.issubdtype(im_arr.dtype, numpy.floating):
            return float(label)
        if numpy.issubdtype(im_arr.dtype, bool):
            label_lower = label.lower()
            if label_lower in ("0", "1", "false", "true"):
                return label_lower in ("1", "true")
            else:
                raise AssertionError(f'Label "{label}" incompatible with label type "{im_arr.dtype}"')
        raise AssertionError(f'Unsupported image label type: "{im_arr.dtype}"')

    # Determine labels present in the image.
    present_labels: List[Any] = numpy.unique(im_arr).tolist()

    # Apply filtering due to `labels` (keep only those).
    if labels is None:
        labels = []
    elif isinstance(labels, str):
        labels = [cast_label(label) for label in labels.split(",") if len(label) > 0]
    if len(labels) > 0:
        present_labels = [label for label in present_labels if label in labels]

    # Apply filtering due to `exclude_labels`.
    if exclude_labels is None:
        exclude_labels = []
    elif isinstance(exclude_labels, str):
        exclude_labels = [cast_label(label) for label in exclude_labels.split(",") if len(label) > 0]
    present_labels = [label for label in present_labels if label not in exclude_labels]

    # Return the image data and the labels.
    return im_arr, present_labels


def assert_has_image_n_labels(
    output_bytes: OutputBytes,
    channel: Channel = None,
    slice: Slice = None,
    frame: Frame = None,
    labels: Labels = None,
    exclude_labels: ExcludeLabels = None,
    n: NumLabels = None,
    delta: NumLabelsDelta = 0,
    min: NumLabelsMin = None,
    max: NumLabelsMax = None,
    negate: Negate = NEGATE_DEFAULT,
) -> None:
    """
    Asserts the output is an image and has the specified labels.

    Labels can be a number of labels or unique values (e.g.,
    ``<has_image_n_labels n="187" exclude_labels="0" />``).

    The primary usage of this assertion is to verify the number of objects in images with uniquely labeled objects.
    """
    present_labels = _get_image_labels(output_bytes, channel, slice, frame, labels, exclude_labels)[1]
    _assert_number(
        len(present_labels),
        n,
        delta,
        min,
        max,
        negate,
        "{expected} labels {n}+-{delta}",
        "{expected} labels to be in [{min}:{max}]",
    )


def assert_has_image_mean_object_size(
    output_bytes: OutputBytes,
    channel: Channel = None,
    slice: Slice = None,
    frame: Frame = None,
    labels: Labels = None,
    exclude_labels: ExcludeLabels = None,
    mean_object_size: MeanObjectSize = None,
    eps: MeanObjectSizeEps = 0.01,
    min: MeanObjectSizeMin = None,
    max: MeanObjectSizeMax = None,
) -> None:
    """
    Asserts the output is an image with labeled objects which have the specified mean size (number of pixels),

    The mean size is plus/minus ``eps`` (e.g., ``<has_image_mean_object_size mean_object_size="111.87" exclude_labels="0" />``).

    The labels must be unique.
    """
    im_arr, present_labels = _get_image_labels(output_bytes, channel, slice, frame, labels, exclude_labels)
    assert (
        im_arr.shape[-1] == 1
    ), f"has_image_mean_object_size is undefined for multi-channel images (channels: {im_arr.shape[-1]})"

    # Build list of all object sizes over all time-frames
    object_sizes = sum(
        [
            # Iterate over all XYZC time-frames (axis C is singleton)
            [(im_arr_t == label).sum() for label in present_labels]
            for im_arr_t in im_arr
        ],
        [],
    )

    # Compute the mean object size and verify
    actual_mean_object_size = numpy.mean([object_size for object_size in object_sizes if object_size > 0])
    _assert_float(
        actual=actual_mean_object_size,
        label="mean object size",
        tolerance=eps,
        expected=mean_object_size,
        range_min=min,
        range_max=max,
    )



from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Type,
    Union,
)

from pydantic import BaseModel
from typing_extensions import Literal

from galaxy.tool_util_models.parameters import (
    create_job_internal_model,
    create_landing_request_internal_model,
    create_landing_request_model,
    create_relaxed_request_model,
    create_request_internal_dereferenced_model,
    create_request_internal_model,
    create_request_model,
    create_test_case_model,
    create_workflow_step_linked_model,
    create_workflow_step_model,
    StateRepresentationT,
    ToolParameterBundle,
    ToolParameterBundleModel,
    ToolParameterT,
)
from .model_validation import (
    validate_against_model,
)

HasToolParameters = Union[List[ToolParameterT], ToolParameterBundle]


class ToolState(ABC):
    input_state: Dict[str, Any]

    def __init__(self, input_state: Dict[str, Any]):
        self.input_state = input_state

    def _validate(self, pydantic_model: Type[BaseModel]) -> None:
        validate_against_model(pydantic_model, self.input_state)

    def validate(self, parameters: HasToolParameters, name: Optional[str] = None) -> None:
        base_model = self.parameter_model_for(parameters, name=name)
        if base_model is None:
            raise NotImplementedError(
                f"Validating tool state against state representation {self.state_representation} is not implemented."
            )
        self._validate(base_model)

    @property
    @abstractmethod
    def state_representation(self) -> StateRepresentationT:
        """Get state representation of the inputs."""

    @classmethod
    def parameter_model_for(cls, parameters: HasToolParameters, name: Optional[str] = None) -> Type[BaseModel]:
        bundle: ToolParameterBundle
        if isinstance(parameters, list):
            bundle = ToolParameterBundleModel(parameters=parameters)
        else:
            bundle = parameters
        return cls._parameter_model_for(bundle, name=name)

    @classmethod
    @abstractmethod
    def _parameter_model_for(cls, parameters: ToolParameterBundle, name: Optional[str] = None) -> Type[BaseModel]:
        """Return a model type for this tool state kind."""


class RelaxedRequestToolState(ToolState):
    state_representation: Literal["relaxed_request"] = "relaxed_request"

    @classmethod
    def _parameter_model_for(cls, parameters: ToolParameterBundle, name: Optional[str] = None) -> Type[BaseModel]:
        return create_relaxed_request_model(parameters, name)


class RequestToolState(ToolState):
    state_representation: Literal["request"] = "request"

    @classmethod
    def _parameter_model_for(cls, parameters: ToolParameterBundle, name: Optional[str] = None) -> Type[BaseModel]:
        return create_request_model(parameters, name)


class RequestInternalToolState(ToolState):
    state_representation: Literal["request_internal"] = "request_internal"

    @classmethod
    def _parameter_model_for(cls, parameters: ToolParameterBundle, name: Optional[str] = None) -> Type[BaseModel]:
        return create_request_internal_model(parameters, name)


class LandingRequestToolState(ToolState):
    state_representation: Literal["landing_request"] = "landing_request"

    @classmethod
    def _parameter_model_for(cls, parameters: ToolParameterBundle, name: Optional[str] = None) -> Type[BaseModel]:
        return create_landing_request_model(parameters, name)


class LandingRequestInternalToolState(ToolState):
    state_representation: Literal["landing_request_internal"] = "landing_request_internal"

    @classmethod
    def _parameter_model_for(cls, parameters: ToolParameterBundle, name: Optional[str] = None) -> Type[BaseModel]:
        return create_landing_request_internal_model(parameters, name)


class RequestInternalDereferencedToolState(ToolState):
    state_representation: Literal["request_internal_dereferenced"] = "request_internal_dereferenced"

    @classmethod
    def _parameter_model_for(cls, parameters: ToolParameterBundle, name: Optional[str] = None) -> Type[BaseModel]:
        return create_request_internal_dereferenced_model(parameters, name)


class JobInternalToolState(ToolState):
    state_representation: Literal["job_internal"] = "job_internal"

    @classmethod
    def _parameter_model_for(cls, parameters: ToolParameterBundle, name: Optional[str] = None) -> Type[BaseModel]:
        return create_job_internal_model(parameters, name)


class TestCaseToolState(ToolState):
    state_representation: Literal["test_case_xml"] = "test_case_xml"

    @classmethod
    def _parameter_model_for(cls, parameters: ToolParameterBundle, name: Optional[str] = None) -> Type[BaseModel]:
        # implement a test case model...
        return create_test_case_model(parameters, name)


class WorkflowStepToolState(ToolState):
    state_representation: Literal["workflow_step"] = "workflow_step"

    @classmethod
    def _parameter_model_for(cls, parameters: ToolParameterBundle, name: Optional[str] = None) -> Type[BaseModel]:
        return create_workflow_step_model(parameters, name)


class WorkflowStepLinkedToolState(ToolState):
    state_representation: Literal["workflow_step_linked"] = "workflow_step_linked"

    @classmethod
    def _parameter_model_for(cls, parameters: ToolParameterBundle, name: Optional[str] = None) -> Type[BaseModel]:
        return create_workflow_step_linked_model(parameters, name)



import json
from collections.abc import MutableMapping
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
)

import packaging.version

from galaxy.tool_util.deps import requirements
from galaxy.tool_util.parser.util import (
    DEFAULT_DECOMPRESS,
    DEFAULT_DELTA,
    DEFAULT_DELTA_FRAC,
    DEFAULT_SORT,
)
from galaxy.tool_util_models.parameter_validators import AnyValidatorModel
from galaxy.tool_util_models.tool_source import (
    HelpContent,
    XrefDict,
    YamlTemplateConfigFile,
)
from galaxy.util import listify
from .interface import (
    AssertionDict,
    AssertionList,
    InputSource,
    PageSource,
    PagesSource,
    ToolSource,
    ToolSourceTest,
    ToolSourceTests,
)
from .output_actions import ToolOutputActionApp
from .output_collection_def import dataset_collector_descriptions_from_output_dict
from .output_objects import (
    ToolOutput,
    ToolOutputCollection,
    ToolOutputCollectionStructure,
)
from .parameter_validators import parse_dict_validators
from .stdio import error_on_exit_code
from .util import is_dict


class YamlToolSource(ToolSource):
    language = "yaml"

    def __init__(self, root_dict: Dict, source_path=None):
        self.root_dict = root_dict
        self._source_path = source_path
        self._macro_paths: List[str] = []

    @property
    def source_path(self):
        return self._source_path

    def parse_class(self):
        return self.root_dict.get("class")

    def parse_tool_type(self):
        return self.root_dict.get("tool_type")

    def parse_id(self):
        return self.root_dict.get("id")

    def parse_version(self) -> Optional[str]:
        version_raw = self.root_dict.get("version")
        return str(version_raw) if version_raw is not None else None

    def parse_name(self) -> str:
        rval = self.root_dict.get("name") or self.parse_id()
        assert rval
        return str(rval)

    def parse_description(self) -> str:
        return self.root_dict.get("description") or ""

    def parse_icon(self) -> Optional[str]:
        icon_elem = self.root_dict.get("icon", {})
        return icon_elem.get("src") if icon_elem is not None else None

    def parse_edam_operations(self) -> List[str]:
        return self.root_dict.get("edam_operations") or []

    def parse_edam_topics(self) -> List[str]:
        return self.root_dict.get("edam_topics") or []

    def parse_xrefs(self) -> List[XrefDict]:
        xrefs = self.root_dict.get("xrefs") or []
        return [XrefDict(value=xref["value"], type=xref["type"]) for xref in xrefs if xref["type"]]

    def parse_sanitize(self):
        return self.root_dict.get("sanitize", True)

    def parse_display_interface(self, default):
        return self.root_dict.get("display_interface", default)

    def parse_require_login(self, default):
        return self.root_dict.get("require_login", default)

    def parse_command(self):
        return self.root_dict.get("command")

    def parse_expression(self):
        return self.root_dict.get("expression")

    def parse_shell_command(self) -> Optional[str]:
        return self.root_dict.get("shell_command")

    def parse_base_command(self) -> Optional[List[str]]:
        """Return string containing script entrypoint."""
        return listify(self.root_dict.get("base_command"))

    def parse_arguments(self) -> Optional[List[str]]:
        return self.root_dict.get("arguments")

    def parse_environment_variables(self):
        return []

    def parse_template_configfiles(self):
        return [YamlTemplateConfigFile(**config) for config in self.root_dict.get("configfiles") or []]

    def parse_interpreter(self):
        return self.root_dict.get("interpreter")

    def parse_version_command(self):
        return self.root_dict.get("runtime_version", {}).get("command", None)

    def parse_version_command_interpreter(self):
        return self.root_dict.get("runtime_version", {}).get("interpreter", None)

    def parse_requirements(self):
        mixed_requirements = self.root_dict.get("requirements", [])
        container = self.root_dict.get("container")
        containers = self.root_dict.get("containers")
        if container:
            if isinstance(container, str):
                container = {"identifier": container, "type": "docker", "explicit": True}
            containers = [container]
        elif containers:
            containers = containers
        else:
            containers = []
        return requirements.parse_requirements_from_lists(
            software_requirements=[r for r in mixed_requirements if r.get("type") == "package"],
            containers=containers,
            resource_requirements=[r for r in mixed_requirements if r.get("type") == "resource"],
            javascript_requirements=[r for r in mixed_requirements if r.get("type") == "javascript"],
            credentials=self.root_dict.get("credentials", []),
        )

    def parse_input_pages(self) -> PagesSource:
        # All YAML tools have only one page (feature is deprecated)
        page_source = YamlPageSource(self.root_dict.get("inputs", {}))
        return PagesSource([page_source])

    def parse_strict_shell(self):
        # TODO: Add ability to disable this.
        return True

    def parse_stdio(self):
        return error_on_exit_code()

    def parse_help(self) -> Optional[HelpContent]:
        help = self.root_dict.get("help")
        format = "markdown"
        if isinstance(help, dict):
            format = help.get("format", "markdown")
        if isinstance(help, str):
            return HelpContent(format=format, content=help)
        elif help and "content" in help:
            return HelpContent(format=format, content=help["content"])
        else:
            return None

    def parse_outputs(self, app: Optional[ToolOutputActionApp]):
        outputs = self.root_dict.get("outputs", [])
        if isinstance(outputs, MutableMapping):
            for name, output_dict in outputs.items():
                output_dict["name"] = name
            outputs = outputs.values()

        output_defs = []
        output_collection_defs = []
        for output_dict in outputs:
            output_type = output_dict.get("type", "data")
            name = output_dict["name"]
            if output_type == "data":
                output_defs.append(self._parse_output(app, name, output_dict))
            elif output_type == "collection":
                output_collection_defs.append(self._parse_output_collection(app, name, output_dict))
            else:
                message = f"Unknown output_type [{output_type}] encountered."
                raise Exception(message)
        outputs = {}
        for output in output_defs:
            outputs[output.name] = output
        output_collections = {}
        for output in output_collection_defs:
            output_collections[output.name] = output

        return outputs, output_collections

    def _parse_output(self, app, name, output_dict):
        output = ToolOutput.from_dict(name, output_dict, app=app)
        return output

    def _parse_output_collection(self, tool, name, output_dict):
        name = output_dict.get("name")
        label = output_dict.get("label")
        default_format = output_dict.get("format", "data")
        collection_type = output_dict.get("type", None)
        collection_type_source = output_dict.get("type_source", None)
        structured_like = output_dict.get("structured_like", None)
        inherit_format = False
        inherit_metadata = False
        if structured_like:
            inherit_format = output_dict.get("inherit_format", None)
            inherit_metadata = output_dict.get("inherit_metadata", None)
        default_format_source = output_dict.get("format_source", None)
        default_metadata_source = output_dict.get("metadata_source", None)
        filters = []
        dataset_collector_descriptions = dataset_collector_descriptions_from_output_dict(output_dict)

        structure = ToolOutputCollectionStructure(
            collection_type=collection_type,
            collection_type_source=collection_type_source,
            structured_like=structured_like,
            dataset_collector_descriptions=dataset_collector_descriptions,
        )
        output_collection = ToolOutputCollection(
            name,
            structure,
            label=label,
            filters=filters,
            default_format=default_format,
            inherit_format=inherit_format,
            inherit_metadata=inherit_metadata,
            default_format_source=default_format_source,
            default_metadata_source=default_metadata_source,
        )
        return output_collection

    def parse_tests_to_dict(self) -> ToolSourceTests:
        tests: List[ToolSourceTest] = []
        rval: ToolSourceTests = dict(tests=tests)

        for i, test_dict in enumerate(self.root_dict.get("tests", [])):
            tests.append(_parse_test(i, test_dict))

        return rval

    def parse_profile(self) -> str:
        return self.root_dict.get("profile") or "16.04"

    def parse_license(self) -> Optional[str]:
        return self.root_dict.get("license")

    def parse_interactivetool(self):
        return self.root_dict.get("entry_points", [])

    def parse_python_template_version(self):
        python_template_version = self.root_dict.get("python_template_version")
        if python_template_version is not None:
            python_template_version = packaging.version.Version(python_template_version)
        return python_template_version

    def to_string(self):
        # TODO: Unit test for dumping/restoring
        return json.dumps(self.root_dict, ensure_ascii=False, sort_keys=False)


def _parse_test(i, test_dict) -> ToolSourceTest:
    inputs = test_dict["inputs"]
    if is_dict(inputs):
        new_inputs = []
        for key, value in inputs.items():
            new_inputs.append({"name": key, "value": value, "attributes": {}})
        test_dict["inputs"] = new_inputs

    outputs = test_dict["outputs"]

    new_outputs = []
    if is_dict(outputs):
        for key, value in outputs.items():
            if is_dict(value):
                attributes = value
                file = attributes.get("file")
            else:
                file = value
                attributes = {}
            new_outputs.append({"name": key, "value": file, "attributes": attributes})
    else:
        for output in outputs:
            name = output["name"]
            value = output.get("file", None)
            attributes = output
            new_outputs.append({"name": name, "value": value, "attributes": attributes})

    for output in new_outputs:
        attributes = output["attributes"]
        defaults = {
            "compare": "diff",
            "lines_diff": 0,
            "delta": DEFAULT_DELTA,
            "delta_frac": DEFAULT_DELTA_FRAC,
            "sort": DEFAULT_SORT,
            "decompress": DEFAULT_DECOMPRESS,
        }
        # TODO
        attributes["extra_files"] = []
        # TODO
        attributes["metadata"] = {}
        assert_list = __to_test_assert_list(attributes.get("asserts", []))
        attributes["assert_list"] = assert_list
        _ensure_has(attributes, defaults)

    test_dict["outputs"] = new_outputs
    # TODO: implement output collections for YAML tools.
    test_dict["output_collections"] = []
    test_dict["command"] = __to_test_assert_list(test_dict.get("command", []))
    test_dict["stdout"] = __to_test_assert_list(test_dict.get("stdout", []))
    test_dict["stderr"] = __to_test_assert_list(test_dict.get("stderr", []))
    test_dict["expect_exit_code"] = test_dict.get("expect_exit_code", None)
    test_dict["expect_failure"] = test_dict.get("expect_failure", False)
    test_dict["expect_test_failure"] = test_dict.get("expect_test_failure", False)
    return test_dict


def to_test_assert_list(assertions) -> AssertionList:
    assertions = assertions or []

    def expand_dict_form(item):
        key, value = item
        new_value = value.copy()
        new_value["that"] = key
        return new_value

    if is_dict(assertions):
        assertions = map(expand_dict_form, assertions.items())

    assert_list: List[AssertionDict] = []
    for assertion in assertions:
        # TODO: not handling nested assertions correctly,
        # not sure these are used though.
        if "that" not in assertion:
            new_assertion = {}
            for assertion_key, assertion_value in assertion.items():
                new_assertion["that"] = assertion_key
                new_assertion.update(assertion_value)
            assertion = new_assertion
        children = assertion.pop("asserts", assertion.pop("children", []))
        # if there are no nested assertions then children should be []
        # but to_test_assert_list would return None
        if children:
            children = to_test_assert_list(children)
        assert_dict: AssertionDict = dict(
            tag=assertion["that"],
            attributes=assertion,
            children=children,
        )
        assert_list.append(assert_dict)

    return assert_list or None  # XML variant is None if no assertions made


# Planemo depends on this and was never updated unfortunately.
# https://github.com/galaxyproject/planemo/blob/master/planemo/test/_check_output.py
__to_test_assert_list = to_test_assert_list


class YamlPageSource(PageSource):
    def __init__(self, inputs_list):
        self.inputs_list = inputs_list

    def parse_input_sources(self):
        return list(map(YamlInputSource, self.inputs_list))


class YamlInputSource(InputSource):
    def __init__(self, input_dict, trusted: bool = True):
        self.input_dict = input_dict
        self.trusted = trusted

    def get(self, key, default=None):
        return self.input_dict.get(key, default)

    def get_bool(self, key, default):
        return self.input_dict.get(key, default)

    def get_bool_or_none(self, key, default):
        return self.input_dict.get(key, default)

    def parse_input_type(self):
        input_type = self.input_dict["type"]
        if input_type == "repeat":
            return "repeat"
        elif input_type == "conditional":
            return "conditional"
        else:
            return "param"

    def parse_extensions(self):
        extensions = self.input_dict.get("extensions")
        if not extensions:
            extensions = self.get("format", "data").split(",")
        return [ext.strip().lower() for ext in extensions]

    def parse_nested_inputs_source(self):
        assert self.parse_input_type() == "repeat"
        return YamlPageSource(self.input_dict["blocks"])

    def parse_test_input_source(self):
        test_dict = self.input_dict.get("test_parameter", None)
        assert test_dict is not None, "conditional must contain a `test_parameter` definition"
        return YamlInputSource(test_dict)

    def parse_when_input_sources(self):
        input_dict = self.input_dict

        sources = []
        if "when" in input_dict:
            for key, value in input_dict["when"].items():
                # casting to string because default value for BooleanToolParameter.legal_values is "true" / "false"
                # Unfortunate, but I guess that's ok for now?
                discriminator = "true" if key is True else "false" if key is False else key
                case_page_source = YamlPageSource(value)
                sources.append((discriminator, case_page_source))
        else:
            for value in input_dict.get("whens", []):
                key = value.get("discriminator")
                discriminator = "true" if key is True else "false" if key is False else key
                case_page_source = YamlPageSource(value["parameters"])
                sources.append((discriminator, case_page_source))
        return sources

    def parse_validators(self) -> List[AnyValidatorModel]:
        return parse_dict_validators(self.input_dict.get("validators", []), trusted=self.trusted)

    def parse_static_options(self) -> List[Tuple[str, str, bool]]:
        static_options = []
        input_dict = self.input_dict
        for option in input_dict.get("options", {}):
            value = option.get("value")
            label = option.get("label", value)
            selected = option.get("selected", False)
            static_options.append((label, value, selected))
        return static_options

    def parse_default(self) -> Optional[Dict[str, Any]]:
        input_dict = self.input_dict
        default_def = input_dict.get("default", None)
        return default_def


def _ensure_has(dict, defaults):
    for key, value in defaults.items():
        if key not in dict:
            dict[key] = value



"""This module contains a linting function for a tool's help."""

from typing import (
    TYPE_CHECKING,
    Union,
)

from galaxy.tool_util.lint import Linter
from galaxy.util import (
    rst_to_html,
    unicodify,
)

if TYPE_CHECKING:
    from galaxy.tool_util.lint import LintContext
    from galaxy.tool_util.parser.interface import ToolSource


class HelpMissing(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        root = tool_xml.find("./help")
        if root is None:
            root = tool_xml.getroot()
        help = tool_xml.find("./help")
        if help is None:
            lint_ctx.warn(
                "No help section found, consider adding a help section to your tool.", linter=cls.name(), node=root
            )


class HelpEmpty(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        help = tool_xml.find("./help")
        if help is None:
            return
        help_text = help.text or ""
        if not help_text.strip():
            lint_ctx.warn("Help section appears to be empty.", linter=cls.name(), node=help)


class HelpPresent(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        help = tool_xml.find("./help")
        if help is None:
            return
        help_text = help.text or ""
        if help_text.strip():
            lint_ctx.valid("Tool contains help section.", linter=cls.name(), node=help)


class HelpTODO(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        help = tool_xml.find("./help")
        if help is None:
            return
        help_text = help.text or ""
        if "TODO" in help_text:
            lint_ctx.warn("Help contains TODO text.", linter=cls.name(), node=help)


class HelpInvalidRST(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        help = tool_xml.find("./help")
        if help is None:
            return
        help_text = help.text or ""
        if not help_text.strip():
            return
        invalid_rst = rst_invalid(help_text)
        if invalid_rst:
            lint_ctx.warn(f"Invalid reStructuredText found in help - [{invalid_rst}].", linter=cls.name(), node=help)


class HelpValidRST(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        help = tool_xml.find("./help")
        if help is None:
            return
        help_text = help.text or ""
        if not help_text.strip():
            return
        invalid_rst = rst_invalid(help_text)
        if not invalid_rst:
            lint_ctx.valid("Help contains valid reStructuredText.", linter=cls.name(), node=help)


def rst_invalid(text: str) -> Union[bool, str]:
    """
    Predicate to determine if text is invalid reStructuredText.
    Return False if the supplied text is valid reStructuredText or
    a string indicating the problem.
    """
    invalid_rst: Union[bool, str] = False
    try:
        rst_to_html(text, error=True)
    except Exception as e:
        invalid_rst = unicodify(e)
    return invalid_rst



import os
from typing import TYPE_CHECKING

from galaxy.tool_util.lint import Linter
from galaxy.tool_util.parser.interface import RequiredFiles

if TYPE_CHECKING:
    from galaxy.tool_util.lint import LintContext
    from galaxy.tool_util.parser import ToolSource


class RequiredFilesExist(Linter):
    """Check that required_files include patterns match existing files."""

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        if not tool_source.source_path:
            return
        tool_dir = os.path.dirname(tool_source.source_path)
        required_files = tool_source.parse_required_files()
        if required_files is None:
            return
        for include in required_files.includes:
            per_include = RequiredFiles.from_dict(
                {
                    "includes": [include],
                    "excludes": [],
                    "extend_default_excludes": False,
                }
            )
            if not per_include.find_required_files(tool_dir):
                path = include["path"]
                path_type = include.get("path_type", "literal")
                if path_type == "literal":
                    lint_ctx.error(
                        f"Required file [{path}] does not exist",
                        linter=cls.name(),
                    )
                else:
                    lint_ctx.error(
                        f"Required files pattern [{path}] (type {path_type}) does not match any files",
                        linter=cls.name(),
                    )



"""This module define an abstract class for reasoning about Galaxy's
dataset collection after jobs are finished.
"""

import abc
from typing import (
    List,
    Optional,
)

from galaxy.tool_util_models.tool_outputs import (
    DatasetCollectionDescriptionT,
    DiscoverViaT,
    FilePatternDatasetCollectionDescription as FilePatternDatasetCollectionDescriptionModel,
    SortCompT,
    SortKeyT,
    ToolProvidedMetadataDatasetCollection as ToolProvidedMetadataDatasetCollectionModel,
)
from galaxy.util import asbool
from .util import is_dict

DEFAULT_EXTRA_FILENAME_PATTERN = (
    r"primary_DATASET_ID_(?P<designation>[^_]+)_(?P<visible>[^_]+)_(?P<ext>[^_]+)(_(?P<dbkey>[^_]+))?"
)
DEFAULT_SORT_BY = "filename"
DEFAULT_SORT_COMP = "lexical"


# XML can describe custom patterns, but these literals describe named
# patterns that will be replaced.
NAMED_PATTERNS = {
    "__default__": DEFAULT_EXTRA_FILENAME_PATTERN,
    "__name__": r"(?P<name>.*)",
    "__designation__": r"(?P<designation>.*)",
    "__name_and_ext__": r"(?P<name>.*)\.(?P<ext>[^\.]+)?",
    "__designation_and_ext__": r"(?P<designation>.*)\.(?P<ext>[^\._]+)?",
}

INPUT_DBKEY_TOKEN = "__input__"
LEGACY_DEFAULT_DBKEY = None  # don't use __input__ for legacy default collection


def dataset_collector_descriptions_from_elem(elem, legacy=True):
    primary_dataset_elems = elem.findall("discover_datasets")
    num_discover_dataset_blocks = len(primary_dataset_elems)
    if num_discover_dataset_blocks == 0 and legacy:
        collectors = [DEFAULT_DATASET_COLLECTOR_DESCRIPTION]
    else:
        default_format = elem.attrib.get("format")
        collectors = []
        for e in primary_dataset_elems:
            description_attributes = e.attrib
            if default_format and "format" not in description_attributes and "ext" not in description_attributes:
                description_attributes["format"] = default_format
            collectors.append(dataset_collection_description(**description_attributes))

    return _validate_collectors(collectors)


def dataset_collector_descriptions_from_output_dict(as_dict):
    discover_datasets_dicts = as_dict.get("discover_datasets") or []
    if is_dict(discover_datasets_dicts):
        discover_datasets_dicts = [discover_datasets_dicts]
    dataset_collector_descriptions = dataset_collector_descriptions_from_list(discover_datasets_dicts)
    return _validate_collectors(dataset_collector_descriptions)


def _validate_collectors(collectors):
    num_discover_dataset_blocks = len(collectors)
    if num_discover_dataset_blocks > 1:
        for collector in collectors:
            if collector.discover_via == "tool_provided_metadata":
                raise Exception(
                    "Cannot specify more than one discover dataset condition if any of them specify tool_provided_metadata."
                )

    return collectors


def dataset_collector_descriptions_from_list(discover_datasets_dicts):
    return [dataset_collection_description(**kwds) for kwds in discover_datasets_dicts]


def dataset_collection_description(**kwargs):
    from_provided_metadata = asbool(kwargs.get("from_provided_metadata", False))
    discover_via = kwargs.get("discover_via", "tool_provided_metadata" if from_provided_metadata else "pattern")
    if discover_via == "tool_provided_metadata":
        for key in ["pattern", "sort_by"]:
            if kwargs.get(key):
                raise Exception(f"Cannot specify attribute [{key}] if from_provided_metadata is True")
        return ToolProvidedMetadataDatasetCollection(**kwargs)
    else:
        return FilePatternDatasetCollectionDescription(**kwargs)


class DatasetCollectionDescription(metaclass=abc.ABCMeta):
    discover_via: DiscoverViaT
    default_ext: Optional[str]
    default_visible: bool
    assign_primary_output: bool
    directory: Optional[str]
    recurse: bool
    match_relative_path: bool

    def __init__(self, **kwargs):
        self.default_dbkey = kwargs.get("dbkey", INPUT_DBKEY_TOKEN)
        self.default_ext = kwargs.get("ext", None)
        if self.default_ext is None and "format" in kwargs:
            self.default_ext = kwargs.get("format")
        self.default_visible = asbool(kwargs.get("visible", None))
        self.assign_primary_output = asbool(kwargs.get("assign_primary_output", False))
        self.directory = kwargs.get("directory", None)
        self.recurse = False
        self.match_relative_path = asbool(kwargs.get("match_relative_path", False))

    def _common_model_props(self):
        return {
            "discover_via": self.discover_via,
            "dbkey": self.default_dbkey,
            "format": self.default_ext,
            "visible": self.default_visible,
            "assign_primary_output": self.assign_primary_output,
            "directory": self.directory,
            "recurse": self.recurse,
            "match_relative_path": self.match_relative_path,
        }

    @abc.abstractmethod
    def to_model(self) -> DatasetCollectionDescriptionT: ...

    def to_dict(self) -> dict:
        return self.to_model().model_dump()

    @property
    def discover_patterns(self) -> List[str]:
        return []


class ToolProvidedMetadataDatasetCollection(DatasetCollectionDescription):
    discover_via = "tool_provided_metadata"

    def to_model(self) -> ToolProvidedMetadataDatasetCollectionModel:
        return ToolProvidedMetadataDatasetCollectionModel(
            discover_via=self.discover_via,
            dbkey=self.default_dbkey,
            format=self.default_ext,
            visible=self.default_visible,
            assign_primary_output=self.assign_primary_output,
            directory=self.directory,
            recurse=self.recurse,
            match_relative_path=self.match_relative_path,
        )

    def to_dict(self) -> dict:
        return self.to_model().model_dump()


class FilePatternDatasetCollectionDescription(DatasetCollectionDescription):
    discover_via = "pattern"
    sort_key: SortKeyT
    sort_comp: SortCompT
    sort_reverse: bool
    pattern: str

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        pattern = kwargs.get("pattern", "__default__")
        self.recurse = asbool(kwargs.get("recurse", False))
        self.match_relative_path = asbool(kwargs.get("match_relative_path", False))
        if pattern in NAMED_PATTERNS:
            pattern = NAMED_PATTERNS[pattern]
        self.pattern = pattern
        if "sort_by" not in kwargs and "sort_key" in kwargs and "sort_comp" in kwargs and "sort_reverse" in kwargs:
            self.sort_reverse = kwargs["sort_reverse"]
            self.sort_comp = kwargs["sort_comp"]
            self.sort_key = kwargs["sort_key"]
        else:
            self.sort_by = sort_by = kwargs.get("sort_by", DEFAULT_SORT_BY)
            if sort_by.startswith("reverse_"):
                self.sort_reverse = True
                sort_by = sort_by[len("reverse_") :]
            else:
                self.sort_reverse = False
            if "_" in sort_by:
                sort_comp, sort_by = sort_by.split("_", 1)
                assert sort_comp in ["lexical", "numeric"]
            else:
                sort_comp = DEFAULT_SORT_COMP
            assert sort_by in ["filename", "name", "designation", "dbkey"]
            self.sort_key = sort_by
            self.sort_comp = sort_comp

    def to_model(self) -> FilePatternDatasetCollectionDescriptionModel:
        return FilePatternDatasetCollectionDescriptionModel(
            discover_via=self.discover_via,
            dbkey=self.default_dbkey,
            format=self.default_ext,
            visible=self.default_visible,
            assign_primary_output=self.assign_primary_output,
            directory=self.directory,
            recurse=self.recurse,
            match_relative_path=self.match_relative_path,
            sort_key=self.sort_key,
            sort_comp=self.sort_comp,
            pattern=self.pattern,
            sort_by=self.sort_by,
            sort_reverse=self.sort_reverse,
        )

    @property
    def discover_patterns(self) -> List[str]:
        return [self.pattern]


DEFAULT_DATASET_COLLECTOR_DESCRIPTION = FilePatternDatasetCollectionDescription(
    default_dbkey=LEGACY_DEFAULT_DBKEY,
)



from typing import (
    Any,
    List,
    Optional,
    Union,
)

from typing_extensions import (
    Annotated,
    Protocol,
)


class AssertionParameter:
    doc: str
    xml_type: Optional[str]
    json_type: Optional[str]
    deprecated: bool
    validators: List[str]

    def __init__(
        self,
        doc: Optional[str],
        xml_type: Optional[str] = None,
        json_type: Optional[str] = None,
        deprecated: bool = False,
        validators: Optional[List[str]] = None,
    ):
        self.doc = doc or ""
        self.xml_type = xml_type
        self.json_type = json_type
        self.deprecated = deprecated
        self.validators = validators or []


XmlInt = Union[int, str]
XmlFloat = Union[float, str]
XmlBool = Union[bool, str]
XmlRegex = str
OptionalXmlInt = Optional[XmlInt]
OptionalXmlFloat = Optional[XmlFloat]
OptionalXmlBool = Optional[XmlBool]

Output = Annotated[str, "The target output of a tool or workflow read as a UTF-8 string"]
OutputBytes = Annotated[bytes, "The target output of a tool or workflow read as raw Python 'bytes'"]


class VerifyAssertionsFunction(Protocol):

    def __call__(self, data: bytes, assertion_description_list: list, decompress: bool = False):
        """Callback for recursirve functions."""


ChildAssertions = Annotated[Any, "Parsed child assertions"]
Negate = Annotated[
    XmlBool,
    AssertionParameter(
        "A boolean that can be set to true to negate the outcome of the assertion.", xml_type="PermissiveBoolean"
    ),
]
NEGATE_DEFAULT = False

N = Annotated[
    Optional[XmlInt], AssertionParameter("Desired number, can be suffixed by ``(k|M|G|T|P|E)i?``", xml_type="Bytes")
]
Delta = Annotated[
    XmlInt,
    AssertionParameter(
        "Allowed difference with respect to n (default: 0), can be suffixed by ``(k|M|G|T|P|E)i?``", xml_type="Bytes"
    ),
]
Min = Annotated[
    Optional[XmlInt],
    AssertionParameter("Minimum number (default: -infinity), can be suffixed by ``(k|M|G|T|P|E)i?``", xml_type="Bytes"),
]
Max = Annotated[
    Optional[XmlInt],
    AssertionParameter("Maximum number (default: infinity), can be suffixed by ``(k|M|G|T|P|E)i?``", xml_type="Bytes"),
]


__all__ = (
    "Annotated",
    "AssertionParameter",
    "ChildAssertions",
    "Delta",
    "Max",
    "Min",
    "Negate",
    "N",
    "NEGATE_DEFAULT",
    "OptionalXmlBool",
    "OptionalXmlFloat",
    "OptionalXmlInt",
    "Output",
    "OutputBytes",
    "VerifyAssertionsFunction",
    "XmlBool",
    "XmlFloat",
    "XmlInt",
    "XmlRegex",
)



"""Utilities for converting between request states."""

import logging
from copy import deepcopy
from typing import (
    Any,
    Callable,
    cast,
    Dict,
    List,
    Optional,
)

from galaxy.tool_util_models.parameters import (
    BooleanParameterModel,
    ConditionalParameterModel,
    ConditionalWhen,
    create_job_runtime_model,
    DataCollectionParameterModel,
    DataCollectionRequest,
    DataColumnParameterModel,
    DataParameterModel,
    DataRequestCollectionUri,
    DataRequestHda,
    DataRequestInternalHda,
    DataRequestInternalHdca,
    DataRequestUri,
    DiscriminatorType,
    DrillDownParameterModel,
    FloatParameterModel,
    GenomeBuildParameterModel,
    HiddenParameterModel,
    IntegerParameterModel,
    RepeatParameterModel,
    SectionParameterModel,
    SelectParameterModel,
    TextParameterModel,
    ToolParameterBundle,
    ToolParameterT,
)
from galaxy.tool_util_models.tool_source import (
    JsonTestCollectionDefDict,
    JsonTestDatasetDefDict,
)
from .state import (
    JobInternalToolState,
    LandingRequestInternalToolState,
    LandingRequestToolState,
    RelaxedRequestToolState,
    RequestInternalDereferencedToolState,
    RequestInternalToolState,
    RequestToolState,
    TestCaseToolState,
)
from .visitor import (
    Callback,
    validate_explicit_conditional_test_value,
    visit_input_values,
    VISITOR_NO_REPLACEMENT,
)

log = logging.getLogger(__name__)


DecodeFunctionT = Callable[[str], int]
EncodeFunctionT = Callable[[int], str]
DereferenceCallable = Callable[[DataRequestUri], DataRequestInternalHda]
DereferenceCollectionCallable = Callable[[DataRequestCollectionUri], DataRequestInternalHdca]
# interfaces for adapting test data dictionaries to tool request dictionaries
# e.g. {class: File, path: foo.bed} => {src: hda, id: ab1235cdfea3}
AdaptDatasets = Callable[[JsonTestDatasetDefDict], DataRequestHda]
AdaptCollections = Callable[[JsonTestCollectionDefDict], DataCollectionRequest]

OPENAPI_REF_TEMPLATE = "#/components/schemas/{model}"


def cwl_runtime_model(input_models: ToolParameterBundle):
    model = create_job_runtime_model(input_models)
    openapi_schema: Dict[str, Any] = {
        "openapi": "3.1.0",
        "info": {
            "title": "Custom API",
            "version": "1.0.0",
            "description": "Generated by Pydantic with OpenAPI customization.",
        },
        "paths": {},  # Define paths manually or generate them elsewhere
        "components": {},
    }
    schemas = model.model_json_schema(mode="serialization", ref_template=OPENAPI_REF_TEMPLATE)
    openapi_schema["components"]["schemas"] = schemas.pop("$defs", {})
    openapi_schema["components"]["schemas"]["inputs"] = schemas
    return openapi_schema


def decode(
    external_state: RequestToolState,
    input_models: ToolParameterBundle,
    decode_id: Callable[[str], int],
    name_base: Optional[str] = None,
) -> RequestInternalToolState:
    """Prepare an internal representation of tool state (request_internal) for storing in the database."""

    external_state.validate(input_models, f"{name_base} (request model)")
    decode_callback = _decode_callback_for(decode_id)
    internal_state_dict = visit_input_values(
        input_models,
        external_state,
        decode_callback,
    )

    internal_request_state = RequestInternalToolState(internal_state_dict)
    internal_request_state.validate(input_models, f"{name_base} (request internal model)")
    return internal_request_state


def encode(
    internal_state: RequestInternalToolState, input_models: ToolParameterBundle, encode_id: EncodeFunctionT
) -> RequestToolState:
    """Prepare an external representation of tool state (request) from persisted state in the database (request_internal)."""

    encode_callback = _encode_callback_for(encode_id)
    request_state_dict = visit_input_values(
        input_models,
        internal_state,
        encode_callback,
    )
    request_state = RequestToolState(request_state_dict)
    request_state.validate(input_models)
    return request_state


def landing_decode(
    external_state: LandingRequestToolState, input_models: ToolParameterBundle, decode_id: Callable[[str], int]
) -> LandingRequestInternalToolState:
    """Prepare an external representation of tool state (request) for storing in the database (request_internal)."""

    external_state.validate(input_models)
    decode_callback = _decode_callback_for(decode_id)
    internal_state_dict = visit_input_values(
        input_models,
        external_state,
        decode_callback,
    )

    internal_request_state = LandingRequestInternalToolState(internal_state_dict)
    internal_request_state.validate(input_models)
    return internal_request_state


def landing_encode(
    internal_state: LandingRequestInternalToolState, input_models: ToolParameterBundle, encode_id: EncodeFunctionT
) -> LandingRequestToolState:
    """Prepare an external representation of tool state (request) for storing in the database (request_internal)."""

    encode_callback = _encode_callback_for(encode_id)
    request_state_dict = visit_input_values(
        input_models,
        internal_state,
        encode_callback,
    )
    request_state = LandingRequestToolState(request_state_dict)
    request_state.validate(input_models)
    return request_state


def strictify(relaxed_state: RelaxedRequestToolState, input_models: ToolParameterBundle) -> RequestToolState:
    """Convert a relaxed request state into a strict request state by applying legacy behavior."""

    tool_state = deepcopy(relaxed_state.input_state)

    def _strictify_parameter(tool_state: Dict[str, Any], parameter: ToolParameterT) -> None:
        if isinstance(parameter, ConditionalParameterModel):
            conditional_state = _initialize_conditional_state(parameter, tool_state)

            test_parameter = parameter.test_parameter
            test_parameter_name = test_parameter.name

            explicit_test_value: Optional[DiscriminatorType] = (
                conditional_state[test_parameter_name] if test_parameter_name in conditional_state else None
            )
            test_value = validate_explicit_conditional_test_value(test_parameter_name, explicit_test_value)
            when = _select_which_when(parameter, test_value, conditional_state)
            _strictify_parameter(conditional_state, test_parameter)
            _strictify_parameters(conditional_state, when)
        elif isinstance(parameter, RepeatParameterModel):
            repeat_instances = _initialize_repeat_state(parameter, tool_state)
            for instance_state in repeat_instances:
                _strictify_parameters(instance_state, parameter)
        elif isinstance(parameter, SectionParameterModel):
            section_state = _initialize_section_state(parameter, tool_state)
            _fill_defaults(section_state, parameter)
        elif isinstance(parameter, TextParameterModel):
            parameter_name = parameter.name
            if parameter_name not in tool_state:
                if not parameter.optional:
                    # restore legacy behavior of allowing empty string implicit default
                    # for these non-optional inputs.
                    tool_state[parameter_name] = parameter.default_value if parameter.default_value is not None else ""
                else:
                    tool_state[parameter_name] = parameter.default_value
            else:
                # legacy behavior of converting explicit None into implicit null. We should introduce
                # a layer somewhere to deal with this behavior further up the stack and clean up these models.
                if not parameter.optional and tool_state[parameter_name] is None:
                    tool_state[parameter_name] = parameter.default_value if parameter.default_value is not None else ""

    def _strictify_parameters(tool_state: Dict[str, Any], input_models: ToolParameterBundle) -> None:
        for parameter in input_models.parameters:
            _strictify_parameter(tool_state, parameter)

    _strictify_parameters(tool_state, input_models)

    request_state = RequestToolState(tool_state)
    request_state.validate(input_models)
    return request_state


def dereference(
    internal_state: RequestInternalToolState,
    input_models: ToolParameterBundle,
    dereference: DereferenceCallable,
    dereference_collection: DereferenceCollectionCallable,
) -> RequestInternalDereferencedToolState:

    def dereference_dict(src_dict: dict):
        src = src_dict.get("src")
        clazz = src_dict.get("class")
        if src == "url":
            data_request_uri: DataRequestUri = DataRequestUri.model_validate(src_dict)
            data_request_hda: DataRequestInternalHda = dereference(data_request_uri)
            return data_request_hda.model_dump()
        elif clazz == "Collection":
            data_request_collection_from_uri: DataRequestCollectionUri = DataRequestCollectionUri.model_validate(
                src_dict
            )
            data_request_hdca: DataRequestInternalHdca = dereference_collection(data_request_collection_from_uri)
            return data_request_hdca.model_dump()
        else:
            return src_dict

    def dereference_callback(parameter: ToolParameterT, value: Any):
        if isinstance(parameter, DataParameterModel):
            if value is None:
                return VISITOR_NO_REPLACEMENT
            if parameter.multiple and isinstance(value, list):
                return list(map(dereference_dict, value))
            else:
                assert isinstance(value, dict), str(value)
                return dereference_dict(value)
        elif isinstance(parameter, DataCollectionParameterModel):
            if value is None:
                return VISITOR_NO_REPLACEMENT
            assert isinstance(value, dict), str(value)
            return dereference_dict(value)
        else:
            return VISITOR_NO_REPLACEMENT

    request_state_dict = visit_input_values(
        input_models,
        internal_state,
        dereference_callback,
    )
    request_state = RequestInternalDereferencedToolState(request_state_dict)
    request_state.validate(input_models)
    return request_state


def encode_test(
    test_case_state: TestCaseToolState,
    input_models: ToolParameterBundle,
    adapt_datasets: AdaptDatasets,
    adapt_collections: AdaptCollections,
):

    def encode_callback(parameter: ToolParameterT, value: Any):
        if isinstance(parameter, DataParameterModel):
            if value is not None:
                if parameter.multiple:
                    assert isinstance(value, list), str(value)
                    test_datasets = cast(List[JsonTestDatasetDefDict], value)
                    return [d.model_dump() for d in map(adapt_datasets, test_datasets)]
                else:
                    assert isinstance(value, dict), str(value)
                    test_dataset = cast(JsonTestDatasetDefDict, value)
                    return adapt_datasets(test_dataset).model_dump()
        elif isinstance(parameter, DataCollectionParameterModel):
            if value is not None:
                assert isinstance(value, dict), str(value)
                test_collection = cast(JsonTestCollectionDefDict, value)
                return adapt_collections(test_collection).model_dump()
        elif isinstance(parameter, SelectParameterModel):
            if parameter.multiple and value is not None and isinstance(value, (str,)):
                return [v.strip() for v in value.split(",")]
            else:
                return VISITOR_NO_REPLACEMENT
        elif isinstance(parameter, DrillDownParameterModel):
            if parameter.multiple and value is not None:
                return [v.strip() for v in value.split(",")]
            else:
                return VISITOR_NO_REPLACEMENT
        elif isinstance(parameter, DataColumnParameterModel):
            if parameter.multiple and value is not None and isinstance(value, (str,)):
                return [int(v.strip()) for v in value.split(",")]
            else:
                return VISITOR_NO_REPLACEMENT

        return VISITOR_NO_REPLACEMENT

    request_state_dict = visit_input_values(
        input_models,
        test_case_state,
        encode_callback,
    )
    request_state = RequestToolState(request_state_dict)
    request_state.validate(input_models)
    return request_state


def fill_static_defaults(
    tool_state: Dict[str, Any], input_models: ToolParameterBundle, profile: float, partial: bool = True
) -> Dict[str, Any]:
    """If additional defaults might stem from Galaxy runtime, partial should be true.

    Setting partial to True, prevents runtime validation.
    """
    _fill_defaults(tool_state, input_models)

    if not partial:
        internal_state = JobInternalToolState(tool_state)
        internal_state.validate(input_models)
    return tool_state


def _fill_defaults(tool_state: Dict[str, Any], input_models: ToolParameterBundle) -> None:
    for parameter in input_models.parameters:
        _fill_default_for(tool_state, parameter)


def _fill_default_for(tool_state: Dict[str, Any], parameter: ToolParameterT) -> None:
    parameter_name = parameter.name
    if isinstance(parameter, BooleanParameterModel):
        if parameter_name not in tool_state:
            # even optional parameters default to false if not in the body of the request :_(
            # see test_tools.py -> expression_null_handling_boolean or test cases for gx_boolean_optional.xml
            tool_state[parameter_name] = parameter.value or False

    if isinstance(parameter, (IntegerParameterModel, FloatParameterModel, HiddenParameterModel)):
        if parameter_name not in tool_state:
            tool_state[parameter_name] = parameter.value
    elif isinstance(parameter, GenomeBuildParameterModel):
        if parameter_name not in tool_state and parameter.optional:
            tool_state[parameter_name] = None
    elif isinstance(parameter, SelectParameterModel):
        # don't fill in dynamic parameters - wait for runtime to specify the default
        if parameter.dynamic_options:
            return

        if parameter_name not in tool_state:
            if not parameter.multiple:
                tool_state[parameter_name] = parameter.default_value
            else:
                tool_state[parameter_name] = parameter.default_values
    elif isinstance(parameter, DrillDownParameterModel):
        if parameter_name not in tool_state:
            if parameter.multiple:
                options = parameter.default_options
                if options is not None:
                    tool_state[parameter_name] = options
            else:
                option = parameter.default_option
                if option is not None:
                    tool_state[parameter_name] = option
    elif isinstance(parameter, ConditionalParameterModel):
        conditional_state = _initialize_conditional_state(parameter, tool_state)

        test_parameter = parameter.test_parameter
        test_parameter_name = test_parameter.name

        explicit_test_value: Optional[DiscriminatorType] = (
            conditional_state[test_parameter_name] if test_parameter_name in conditional_state else None
        )
        test_value = validate_explicit_conditional_test_value(test_parameter_name, explicit_test_value)
        when = _select_which_when(parameter, test_value, conditional_state)
        _fill_default_for(conditional_state, test_parameter)
        _fill_defaults(conditional_state, when)
    elif isinstance(parameter, RepeatParameterModel):
        repeat_instances = _initialize_repeat_state(parameter, tool_state)
        for instance_state in repeat_instances:
            _fill_defaults(instance_state, parameter)
    elif isinstance(parameter, SectionParameterModel):
        section_state = _initialize_section_state(parameter, tool_state)
        _fill_defaults(section_state, parameter)
    elif isinstance(parameter, DataCollectionParameterModel):
        collection_parameter = parameter
        if parameter_name not in tool_state and collection_parameter.optional:
            tool_state[parameter_name] = None
    elif isinstance(parameter, TextParameterModel):
        if parameter_name not in tool_state:
            if not parameter.optional:
                # restore legacy behavior of allowing empty string implicit default
                # for these non-optional inputs.
                tool_state[parameter_name] = parameter.default_value if parameter.default_value is not None else ""
            else:
                tool_state[parameter_name] = parameter.default_value
        else:
            # legacy behavior of converting explicit None into implicit null. We should introduce
            # a layer somewhere to deal with this behavior further up the stack and clean up these models.
            if not parameter.optional and tool_state[parameter_name] is None:
                tool_state[parameter_name] = parameter.default_value if parameter.default_value is not None else ""


def _initialize_section_state(parameter: SectionParameterModel, tool_state: Dict[str, Any]) -> Dict[str, Any]:
    parameter_name = parameter.name
    if parameter_name not in tool_state:
        tool_state[parameter_name] = {}
    section_state = cast(Dict[str, Any], tool_state[parameter_name])
    return section_state


def _initialize_conditional_state(parameter: ConditionalParameterModel, tool_state: Dict[str, Any]) -> Dict[str, Any]:
    parameter_name = parameter.name
    if parameter_name not in tool_state:
        tool_state[parameter_name] = {}

    raw_conditional_state = tool_state[parameter_name]
    assert isinstance(raw_conditional_state, dict)
    conditional_state = cast(Dict[str, Any], raw_conditional_state)
    return conditional_state


def _initialize_repeat_state(parameter: RepeatParameterModel, tool_state: Dict[str, Any]) -> List[Dict[str, Any]]:
    parameter_name = parameter.name
    if parameter_name not in tool_state:
        tool_state[parameter_name] = []
    repeat_instances = cast(List[Dict[str, Any]], tool_state[parameter_name])
    if parameter.min:
        while len(repeat_instances) < parameter.min:
            repeat_instances.append({})
    return repeat_instances


def _select_which_when(
    conditional: ConditionalParameterModel, test_value: Optional[DiscriminatorType], conditional_state: Dict[str, Any]
) -> ConditionalWhen:
    for when in conditional.whens:
        if test_value is None and when.is_default_when:
            return when
        elif test_value == when.discriminator:
            return when
    else:
        raise Exception(
            f"Invalid conditional test value ({test_value}) for parameter ({conditional.test_parameter.name})"
        )


def _encode_callback_for(encode_id: EncodeFunctionT) -> Callback:

    def encode_src_dict(src_dict: dict):
        if "id" in src_dict:
            encoded_dict = src_dict.copy()
            encoded_dict["id"] = encode_id(src_dict["id"])
            return encoded_dict
        else:
            return src_dict

    def encode_element(element: dict):
        if element.get("__class__") == "Batch":
            encoded = element.copy()
            values = encoded.pop("values")
            encoded["values"] = list(map(encode_src_dict, values))
            return encoded
        else:
            return encode_src_dict(element)

    def encode_callback(parameter: ToolParameterT, value: Any):
        if isinstance(parameter, DataParameterModel):
            if parameter.multiple and isinstance(value, list):
                return list(map(encode_element, value))
            else:
                assert isinstance(value, dict), str(value)
                return encode_element(value)
        elif isinstance(parameter, DataCollectionParameterModel):
            assert isinstance(value, dict), str(value)
            return encode_element(value)
        else:
            return VISITOR_NO_REPLACEMENT

    return encode_callback


def _decode_callback_for(decode_id: DecodeFunctionT) -> Callback:

    def decode_src_dict(src_dict: dict):
        if "id" in src_dict:
            decoded_dict = src_dict.copy()
            decoded_dict["id"] = decode_id(src_dict["id"])
            return decoded_dict
        else:
            return src_dict

    def decode_element(element: dict):
        if element.get("__class__") == "Batch":
            decoded = element.copy()
            values = decoded.pop("values")
            decoded["values"] = list(map(decode_src_dict, values))
            return decoded
        else:
            return decode_src_dict(element)

    def decode_callback(parameter: ToolParameterT, value: Any):
        if isinstance(parameter, DataParameterModel):
            if value is None:
                return VISITOR_NO_REPLACEMENT
            if parameter.multiple and isinstance(value, list):
                return list(map(decode_element, value))
            else:
                assert isinstance(value, dict), str(value)
                return decode_element(value)
        elif isinstance(parameter, DataCollectionParameterModel):
            if value is None:
                return VISITOR_NO_REPLACEMENT
            assert isinstance(value, dict), str(value)
            return decode_src_dict(value)
        else:
            return VISITOR_NO_REPLACEMENT

    return decode_callback



import re
from typing import Optional

from lxml.etree import XMLSyntaxError

from galaxy.tool_util.verify import asserts
from galaxy.util import (
    asbool,
    parse_xml_string,
    unicodify,
)
from ._types import (
    Annotated,
    AssertionParameter,
    ChildAssertions,
    Delta,
    Max,
    Min,
    N,
    Negate,
    NEGATE_DEFAULT,
    Output,
    VerifyAssertionsFunction,
    XmlBool,
    XmlRegex,
)
from ._util import _assert_presence_number

Path = Annotated[str, AssertionParameter("The Python xpath-like expression to find the target element.")]
ElementExpression = Annotated[
    XmlRegex,
    AssertionParameter("The regular expressions to apply against the target element.", validators=["check_regex"]),
]
AttributeExpression = Annotated[
    XmlRegex,
    AssertionParameter(
        "The regular expressions to apply against the named attribute on the target XML element.",
        validators=["check_regex"],
    ),
]
Attribute = Annotated[str, AssertionParameter("The XML attribute name to test against from the target XML element.")]
OptionalAttribute = Annotated[
    Optional[str], AssertionParameter("The XML attribute name to test against from the target XML element.")
]
ElementText = Annotated[
    str, AssertionParameter("The expected element text (body of the XML tag) to test against on the target XML element")
]
AttributeText = Annotated[
    str, AssertionParameter("The expected attribute value to test against on the target XML element")
]
All = Annotated[
    XmlBool,
    AssertionParameter(
        "Check the sub-assertions for all paths matching the path. Default: false, i.e. only the first ",
        xml_type="PermissiveBoolean",
    ),
]


def assert_is_valid_xml(output: Output) -> None:
    """Asserts the output is a valid XML file (e.g. ``<is_valid_xml />``)."""
    try:
        parse_xml_string(output)
    except XMLSyntaxError as e:
        raise AssertionError(f"Expected valid XML, but could not parse output. {unicodify(e)}")


def assert_has_element_with_path(output: Output, path: Path, negate: Negate = NEGATE_DEFAULT) -> None:
    """Asserts the XML output contains at least one element (or tag) with the specified
    XPath-like ``path``, e.g.

    ```xml
    <has_element_with_path path="BlastOutput_param/Parameters/Parameters_matrix" />
    ```

    With ``negate`` the result of the assertion can be inverted."""
    assert_xml_element(output, path, negate=negate)


def assert_has_n_elements_with_path(
    output: Output,
    path: Path,
    n: N = None,
    delta: Delta = 0,
    min: Min = None,
    max: Max = None,
    negate: Negate = NEGATE_DEFAULT,
) -> None:
    """Asserts the XML output contains the specified number (``n``, optionally with ``delta``) of elements (or
    tags) with the specified XPath-like ``path``.

    For example:

    ```xml
    <has_n_elements_with_path n="9" path="BlastOutput_iterations/Iteration/Iteration_hits/Hit/Hit_num" />
    ```

    Alternatively to ``n`` and ``delta`` also the ``min`` and ``max`` attributes
    can be used to specify the range of the expected number of occurences.
    With ``negate`` the result of the assertion can be inverted.
    """
    assert_xml_element(output, path, n=n, delta=delta, min=min, max=max, negate=negate)


def assert_element_text_matches(
    output: Output, path: Path, expression: ElementExpression, negate: Negate = NEGATE_DEFAULT
) -> None:
    """Asserts the text of the XML element with the specified XPath-like ``path``
    matches the regular expression defined by ``expression``.

    For example:

    ```xml
    <element_text_matches path="BlastOutput_version" expression="BLASTP\\s+2\\.2.*"/>
    ```

    The assertion implicitly also asserts that an element matching ``path`` exists.
    With ``negate`` the result of the assertion (on the matching) can be inverted (the
    implicit assertion on the existence of the path is not affected).
    """
    sub = {"tag": "has_text_matching", "attributes": {"expression": expression, "negate": negate}}
    assert_xml_element(output, path, asserts.verify_assertions, [sub])


def assert_element_text_is(output: Output, path: Path, text: ElementText, negate: Negate = NEGATE_DEFAULT) -> None:
    """Asserts the text of the XML element with the specified XPath-like ``path`` is
    the specified ``text``.

    For example:

    ```xml
    <element_text_is path="BlastOutput_program" text="blastp" />
    ```

    The assertion implicitly also asserts that an element matching ``path`` exists.
    With ``negate`` the result of the assertion (on the equality) can be inverted (the
    implicit assertion on the existence of the path is not affected).
    """
    assert_element_text_matches(output, path, re.escape(text) + "$", negate=negate)


def assert_attribute_matches(
    output: Output,
    path: Path,
    attribute: Attribute,
    expression: AttributeExpression,
    negate: Negate = NEGATE_DEFAULT,
) -> None:
    """Asserts the XML ``attribute`` for the element (or tag) with the specified
    XPath-like ``path`` matches the regular expression specified by ``expression``.

    For example:

    ```xml
    <attribute_matches path="outerElement/innerElement2" attribute="foo2" expression="bar\\d+" />
    ```

    The assertion implicitly also asserts that an element matching ``path`` exists.
    With ``negate`` the result of the assertion (on the matching) can be inverted (the
    implicit assertion on the existence of the path is not affected).
    """
    sub = {"tag": "has_text_matching", "attributes": {"expression": expression, "negate": negate}}
    assert_xml_element(output, path, asserts.verify_assertions, [sub], attribute=attribute)


def assert_attribute_is(
    output: Output, path: Path, attribute: Attribute, text: AttributeText, negate: Negate = NEGATE_DEFAULT
) -> None:
    """Asserts the XML ``attribute`` for the element (or tag) with the specified
    XPath-like ``path`` is the specified ``text``.

    For example:

    ```xml
    <attribute_is path="outerElement/innerElement1" attribute="foo" text="bar" />
    ```

    The assertion implicitly also asserts that an element matching ``path`` exists.
    With ``negate`` the result of the assertion (on the equality) can be inverted (the
    implicit assertion on the existence of the path is not affected).
    """
    assert_attribute_matches(output, path, attribute, re.escape(text) + "$", negate=negate)


def assert_element_text(
    output: Output,
    path: Path,
    verify_assertions_function: VerifyAssertionsFunction,
    children: ChildAssertions,
    negate: Negate = NEGATE_DEFAULT,
) -> None:
    """This tag allows the developer to recurisively specify additional assertions as
    child elements about just the text contained in the element specified by the
    XPath-like ``path``, e.g.

    ```xml
    <element_text path="BlastOutput_iterations/Iteration/Iteration_hits/Hit/Hit_def">
      <not_has_text text="EDK72998.1" />
    </element_text>
    ```

    The assertion implicitly also asserts that an element matching ``path`` exists.
    With ``negate`` the result of the implicit assertions can be inverted.
    The sub-assertions, which have their own ``negate`` attribute, are not affected
    by ``negate``.
    """
    assert_xml_element(output, path, verify_assertions_function, children, negate=negate)


def assert_xml_element(
    output: Output,
    path: Path,
    verify_assertions_function: Optional[VerifyAssertionsFunction] = None,
    children: ChildAssertions = None,
    attribute: OptionalAttribute = None,
    all: All = False,
    n: N = None,
    delta: Delta = 0,
    min: Min = None,
    max: Max = None,
    negate: Negate = NEGATE_DEFAULT,
) -> None:
    """Assert if the XML file contains element(s) or tag(s) with the specified
    [XPath-like ``path``](https://lxml.de/xpathxslt.html).  If ``n`` and ``delta``
    or ``min`` and ``max`` are given also the number of occurences is checked.

    ```xml
    <assert_contents>
      <xml_element path="./elem"/>
      <xml_element path="./elem/more[2]"/>
      <xml_element path=".//more" n="3" delta="1"/>
    </assert_contents>
    ```

    With ``negate="true"`` the outcome of the assertions wrt the precence and number
    of ``path`` can be negated. If there are any sub assertions then check them against

    - the content of the attribute ``attribute``
    - the element's text if no attribute is given

    ```xml
    <assert_contents>
      <xml_element path="./elem/more[2]" attribute="name">
        <has_text_matching expression="foo$"/>
      </xml_element>
    </assert_contents>
    ```

    Sub-assertions are not subject to the ``negate`` attribute of ``xml_element``.
    If ``all`` is ``true`` then the sub assertions are checked for all occurences.

    Note that all other XML assertions can be expressed by this assertion (Galaxy
    also implements the other assertions by calling this one).
    """
    children = children or []
    all = asbool(all)
    # assert that path is in output (the specified number of times)

    xml = parse_xml_string(output)
    _assert_presence_number(
        xml,
        path,
        n,
        delta,
        min,
        max,
        negate,
        lambda x, p: x.find(p) is not None,
        lambda x, p: len(x.findall(p)),
        "{expected} path '{text}' in xml",
        "{expected} {n}+-{delta} occurrences of path '{text}' in xml",
        "{expected} that the number of occurences of path '{text}' in xml is in [{min}:{max}]",
    )

    # check sub-assertions
    if len(children) == 0 or verify_assertions_function is None:
        return
    for occ in xml.findall(path):
        if attribute is None or attribute == "":
            content = occ.text
        else:
            content = occ.attrib[attribute]  # type: ignore[assignment] # https://github.com/lxml/lxml-stubs/pull/99
        try:
            if content is None:
                raise AssertionError("Failed to find expected XML content")
            verify_assertions_function(content.encode("utf-8"), children)
        except AssertionError as e:
            if attribute is not None and attribute != "":
                raise AssertionError(f"Attribute '{attribute}' on element with path '{path}': {str(e)}")
            else:
                raise AssertionError(f"Text of element with path '{path}': {str(e)}")
        if not all:
            break



import os.path
from os import getenv
from typing import (
    Set,
    TYPE_CHECKING,
    Union,
)

from packaging.version import Version

# from galaxy import config
from galaxy.tool_util.lint import Linter
from galaxy.util import (
    listify,
    parse_xml,
)
from galaxy.util.resources import resource_path

if TYPE_CHECKING:
    from galaxy.tool_util.lint import LintContext
    from galaxy.tool_util.parser import ToolSource
    from galaxy.util.resources import Traversable

DATATYPES_CONF = getenv("DATATYPES_CONF", resource_path(__name__, "datatypes_conf.xml.sample"))


def _parse_datatypes(datatype_conf_path: Union[str, "Traversable"]) -> Set[str]:
    datatypes = set()
    tree = parse_xml(datatype_conf_path)
    root = tree.getroot()
    for elem in root.findall("./registration/datatype"):
        extension = elem.get("extension", "")
        datatypes.add(extension)
        auto_compressed_types = listify(elem.get("auto_compressed_types", ""))
        for act in auto_compressed_types:
            datatypes.add(f"{extension}.{act}")
    return datatypes


class DatatypesCustomConf(Linter):
    """
    Check if a custom datatypes_conf.xml is present
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        if not tool_source.source_path:
            return
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        tool_node = tool_xml.getroot()
        tool_dir = os.path.dirname(tool_source.source_path)
        datatypes_conf_path = os.path.join(tool_dir, "datatypes_conf.xml")
        if os.path.exists(datatypes_conf_path):
            lint_ctx.warn(
                "Tool uses a custom datatypes_conf.xml which is discouraged",
                linter=cls.name(),
                node=tool_node,
            )


class ValidDatatypes(Linter):
    """
    Check that used datatypes are available
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        profile = tool_source.parse_profile()
        # get Galaxy built-in dataypes
        datatypes = _parse_datatypes(DATATYPES_CONF)
        # add custom tool data types
        if tool_source.source_path:
            tool_dir = os.path.dirname(tool_source.source_path)
            datatypes_conf_path = os.path.join(tool_dir, "datatypes_conf.xml")
            if os.path.exists(datatypes_conf_path):
                datatypes |= _parse_datatypes(datatypes_conf_path)
        for attrib in ["format", "ftype", "ext"]:
            for elem in tool_xml.findall(f".//*[@{attrib}]"):
                # skip help section, "format" in help has a different meaning
                if elem.tag == "help":
                    continue
                formats = elem.get(attrib, "").split(",")
                # Certain elements (e.g. `data`) can only have one format. This
                # is checked separately by linting against the XSD.
                if elem.tag == "param":
                    if any(format in ["auto", "input"] for format in formats):
                        lint_ctx.error(
                            "Invalid format (auto or input) in tool or tool test inputs",
                            linter=cls.name(),
                            node=elem,
                        )
                        continue
                for format in formats:
                    if Version(str(profile)) <= Version("16.04") and format == "input":
                        continue
                    if format == "auto":
                        continue
                    if format not in datatypes:
                        lint_ctx.error(
                            f"Unknown datatype [{format}] used in {elem.tag} element",
                            linter=cls.name(),
                            node=elem,
                        )



#!/usr/bin/env python

import argparse
import sys
from json import dumps
from textwrap import (
    indent,
    wrap,
)
from typing import List

from galaxy.tool_util.upgrade import (
    Advice,
    advise_on_upgrade,
    latest_supported_version,
)

LEVEL_TO_STRING = {
    "must_fix": "❌",
    "ready": "✅",
    "consider": "🤔",
    "info": "ℹ️",
}
DESCRIPTION = f"""
A small utility to check for potential problems and provide advice when upgrading a tool's
profile version. This version of the script can provide advice for upgrading tools through
{latest_supported_version}.
"""


def arg_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument("xml_file")
    parser.add_argument(
        "-p",
        "--profile-version",
        dest="profile_version",
        default=latest_supported_version,
        help="Provide upgrade advice up to this Galaxy tool profile version.",
    )
    parser.add_argument(
        "-j",
        "--json",
        default=False,
        action="store_true",
        help="Output aadvice as JSON.",
    )
    parser.add_argument(
        "-n",
        "--niche",
        default=False,
        action="store_true",
        help="Include advice about niche features that may not be relevant for most tools - including the use of 'galaxy.json' and writing global state in the $HOME directory.",
    )
    return parser


def _print_advice(advice: Advice):
    message = "\n".join(wrap(advice.advice_code_message, initial_indent="", subsequent_indent="    "))
    level = advice.level
    level_str = LEVEL_TO_STRING[level]
    url = advice.url
    print(f"- {level_str}{message}\n")
    if advice.message:
        print(indent(advice.message, "    "))
        print("")
    if url:
        print(f"    More information at {url}")


def _print_advice_list(advice_list: List[Advice]):
    for advice in advice_list:
        _print_advice(advice)


def advise(xml_file: str, version: str, json: bool, niche: bool):
    advice_list = advise_on_upgrade(xml_file, version)
    if not niche:
        advice_list = [a for a in advice_list if not a.niche]
    if json:
        print(dumps(a.to_dict() for a in advice_list))
    else:
        _print_advice_list(advice_list)


def main(argv=None) -> None:
    if argv is None:
        argv = sys.argv[1:]

    args = arg_parser().parse_args(argv)
    advise(args.xml_file, args.profile_version, args.json, args.niche)


if __name__ == "__main__":
    main()



import re

from ._types import (
    Annotated,
    AssertionParameter,
    Delta,
    Max,
    Min,
    N,
    Negate,
    NEGATE_DEFAULT,
    Output,
)
from ._util import _assert_number

Sep = Annotated[str, AssertionParameter("Separator defining columns, default: tab")]
Comment = Annotated[
    str,
    AssertionParameter(
        "Comment character(s) used to skip comment lines (which should not be used for counting columns)"
    ),
]


def get_first_line(output: str, comment: str) -> str:
    """
    get the first non-comment and non-empty line
    """
    if comment != "":
        match = re.search(f"^([^{comment}].*)$", output, flags=re.MULTILINE)
    else:
        match = re.search("^(.+)$", output, flags=re.MULTILINE)
    if match is None:
        return ""
    else:
        return match.group(1)


def assert_has_n_columns(
    output: Output,
    n: N = None,
    delta: Delta = 0,
    min: Min = None,
    max: Max = None,
    sep: Sep = "\t",
    comment: Comment = "",
    negate: Negate = NEGATE_DEFAULT,
) -> None:
    """Asserts tabular output  contains the specified
    number (``n``) of columns.

    For instance, ``<has_n_columns n="3"/>``. The assertion tests only the first line.
    Number of columns can optionally also be specified with ``delta``. Alternatively the
    range of expected occurences can be specified by ``min`` and/or ``max``.

    Optionally a column separator (``sep``, default is ``\t``) `and comment character(s)
    can be specified (``comment``, default is empty string). The first non-comment
    line is used for determining the number of columns.
    """
    first_line = get_first_line(output, comment)
    n_columns = len(first_line.split(sep))
    _assert_number(
        n_columns,
        n,
        delta,
        min,
        max,
        negate,
        "{expected} {n}+-{delta} columns in output",
        "{expected} the number of columns in output to be in [{min}:{max}]",
    )



from galaxy.tool_util_models._types import is_optional
from galaxy.tool_util_models.parameters import (
    BooleanParameterModel,
    ColorParameterModel,
    ConditionalParameterModel,
    ConditionalWhen,
    CwlBooleanParameterModel,
    CwlDirectoryParameterModel,
    CwlFileParameterModel,
    CwlFloatParameterModel,
    CwlIntegerParameterModel,
    CwlNullParameterModel,
    CwlStringParameterModel,
    CwlUnionParameterModel,
    DataCollectionParameterModel,
    DataCollectionRequest,
    DataParameterModel,
    DataRequest,
    DataRequestCollectionUri,
    DataRequestHda,
    DataRequestInternalHda,
    DataRequestInternalHdca,
    DataRequestUri,
    FloatParameterModel,
    GalaxyParameterT,
    HiddenParameterModel,
    IntegerParameterModel,
    LabelValue,
    RawStateDict,
    RepeatParameterModel,
    RulesParameterModel,
    SelectParameterModel,
    TextParameterModel,
    ToolParameterBundle,
    ToolParameterBundleModel,
    ToolParameterModel,
    ToolParameterT,
)
from .case import (
    test_case_state,
    validate_test_cases_for_tool_source,
)
from .convert import (
    decode,
    dereference,
    encode,
    encode_test,
    fill_static_defaults,
    landing_decode,
    landing_encode,
    strictify,
)
from .factory import (
    from_input_source,
    input_models_for_pages,
    input_models_for_tool_source,
    input_models_from_json,
    ParameterDefinitionError,
    tool_parameter_bundle_from_json,
)
from .json import to_json_schema_string
from .model_validation import (
    validate_against_model,
    validate_internal_job,
    validate_internal_landing_request,
    validate_internal_request,
    validate_internal_request_dereferenced,
    validate_landing_request,
    validate_relaxed_request,
    validate_request,
    validate_test_case,
    validate_workflow_step,
    validate_workflow_step_linked,
    ValidationFunctionT,
)
from .state import (
    HasToolParameters,
    JobInternalToolState,
    LandingRequestInternalToolState,
    LandingRequestToolState,
    RelaxedRequestToolState,
    RequestInternalDereferencedToolState,
    RequestInternalToolState,
    RequestToolState,
    TestCaseToolState,
    ToolState,
    WorkflowStepLinkedToolState,
    WorkflowStepToolState,
)
from .visitor import (
    flat_state_path,
    keys_starting_with,
    repeat_inputs_to_array,
    validate_explicit_conditional_test_value,
    visit_input_values,
    VISITOR_NO_REPLACEMENT,
)

__all__ = (
    "from_input_source",
    "input_models_for_pages",
    "input_models_for_tool_source",
    "tool_parameter_bundle_from_json",
    "input_models_from_json",
    "ParameterDefinitionError",
    "JobInternalToolState",
    "ToolParameterBundle",
    "ToolParameterBundleModel",
    "DataRequest",
    "DataRequestCollectionUri",
    "DataRequestInternalHda",
    "DataRequestInternalHdca",
    "DataRequestHda",
    "DataRequestUri",
    "DataCollectionRequest",
    "ToolParameterModel",
    "IntegerParameterModel",
    "BooleanParameterModel",
    "CwlFileParameterModel",
    "CwlFloatParameterModel",
    "CwlIntegerParameterModel",
    "CwlStringParameterModel",
    "CwlNullParameterModel",
    "CwlUnionParameterModel",
    "CwlBooleanParameterModel",
    "CwlDirectoryParameterModel",
    "GalaxyParameterT",
    "TextParameterModel",
    "FloatParameterModel",
    "HiddenParameterModel",
    "ColorParameterModel",
    "RulesParameterModel",
    "DataParameterModel",
    "DataCollectionParameterModel",
    "LabelValue",
    "SelectParameterModel",
    "ConditionalParameterModel",
    "ConditionalWhen",
    "RepeatParameterModel",
    "RawStateDict",
    "ValidationFunctionT",
    "validate_against_model",
    "validate_internal_job",
    "validate_internal_landing_request",
    "validate_internal_request",
    "validate_internal_request_dereferenced",
    "validate_landing_request",
    "validate_relaxed_request",
    "validate_request",
    "validate_test_case",
    "validate_workflow_step",
    "validate_workflow_step_linked",
    "validate_explicit_conditional_test_value",
    "is_optional",
    "ToolState",
    "TestCaseToolState",
    "ToolParameterT",
    "HasToolParameters",
    "to_json_schema_string",
    "test_case_state",
    "validate_test_cases_for_tool_source",
    "RelaxedRequestToolState",
    "RequestToolState",
    "RequestInternalToolState",
    "RequestInternalDereferencedToolState",
    "LandingRequestToolState",
    "LandingRequestInternalToolState",
    "flat_state_path",
    "keys_starting_with",
    "visit_input_values",
    "repeat_inputs_to_array",
    "VISITOR_NO_REPLACEMENT",
    "decode",
    "encode",
    "encode_test",
    "fill_static_defaults",
    "landing_decode",
    "landing_encode",
    "dereference",
    "strictify",
    "WorkflowStepToolState",
    "WorkflowStepLinkedToolState",
)
